$TITLE ('COS DCE Boot-mode Main Program, version 01.04')
$PAGELENGTH(42)
;*****************************************************************************
; File -- bootmain.a51
; Purpose -- COS DCE 8051 boot code.
;
;            ------------------------   --------------------------------------
; Authors -- Original code by ...       Adapted for Boot-mode by ...
;            ------------------------   --------------------------------------
;            Mr. Daniel B. Blackman     William I. Clement, Ph.D., President
;            P.O. Box 20627             Clement Engineering, Inc.
;            Oakland, CA  94620         1261 Dogwood Road
;            (510) 910-9047             Arnold, MD  21012-2220
;                                       (410) 518-6591
;            ------------------------   --------------------------------------
;
; Change History --
; 24Aug00   Code version v17088 accepted by WIC from DBB
; 15Sep00   WIC, version 01.00 delivered.
; 29Sep00   WIC, version 01.02 delivered.
; 09Oct00   WIC, version 01.03 delivered.
; 10Oct00   WIC, version 01.04 changes:
;           (1) Set RAM_BANK = '1' in 'safe_controls'
;           (2) Update tests of event flags for each executive loop task.
;           (3) Correct error in 'SETUP_TO_BUILD_HKP' when no diagnostics
;               were received during the current 1-second interval.  In that
;               case, 'MEM_COPY' was told to copy 64KB of RAM.  The zero-
;               diagnostics case is now handled properly. 
; 10Oct00   WIC, version 01.04 delivered.
;*****************************************************************************

$SET (BOOT_MAIN=1)
$INCLUDE(common.inc)

; Comment-out the following line for the flight build ...
;$SET (INCLUDE_TEST_CODE=1)

;*****************************************************************************
;*****************************************************************************
;                         Allocate Internal 8051 RAM
;*****************************************************************************
;*****************************************************************************
   DSEG  at  30h   ; begin immediately after the bit-addressable registers
CTR_TICK:          DS   1   ; counts the Timer0 rollovers
CMDBUFF:           DS   4   ; used for command packet handling
HKPBUFF:           DS   4   ; used while sending packets to CS computer
tempDPTR:          DS   2
maxSP:             DS   1   ; maximum Stack Pointer value
; CRC Cyclic Redundancy Check calculation variables ...
CRCH:              DS   1   ; CRC high byte
CRCL:              DS   1   ; CRC low byte
DXL:               DS   1   ; CRC register
DXH:               DS   1   ; CRC register
TEMP:              DS   1   ; Temporary register
TEMP1:             DS   1   ; Temporary register
; CRC Cyclic Redundancy Check calculation variables (LFDCRC and background)...
MEMPTR:            DS   2   ; pointer to next address for CRC calculation
CRC_LEN:           DS   2   ; number of bytes remaining in background CRC
CRCHa:             DS   1   ; CRCH stored here while process interrupted
CRCLa:             DS   1   ; CRCL stored here while process interrupted
REGION_CRC:        DS   1   ; current CRC region number (for background task)
; Housekeeping Packet background task
SCRIPT_PTR:        DS   2   ; pointer to next housekeeping script entry
HKP_PTR:           DS   2   ; pointer to next housekeeping location to write
HKP_PARAMS:        DS   4   ; holds parameters for multi-item readings

   BSEG  at  00h   ; start at base of bit-addressable registers
; These bits are the event flags which control which tasks may be run
; within the main executive loop and which tasks can interrupt a background
; task ...
RXD_FLAG:          DBIT  1  ; 1 = command opcode received
SECS_FLAG:         DBIT  1  ; 1 = 1-second timer expired
SEND_DUMP_FLAG:    DBIT  1  ; 1 = dump packet queued to send
SEND_HKP_FLAG:     DBIT  1  ; 1 = housekeeping packet queued to send
HKP_BUILD_FLAG:    DBIT  1  ; 1 = building housekeeping packet
TIMER1_DONE:       DBIT  1  ; 1 = Timer1 has timed out
TIMER1_EDGE:       DBIT  1  ; 1 = 'TIMER1_DONE' changed low-to-high
CRC_FLAG:          DBIT  1  ; 1 = background CRC task enabled
; These bits control the operation of various functions ...
SRC_FLAG:          DBIT  1  ; memory source for copy (0=XDATA, 1=CODE)
REGION_FLAG:       DBIT  1  ; memory source for backgnd CRC (0=XDATA, 1=CODE)
CALC_CRC:          DBIT  1  ; 1 = calculate CRC for upload packet
NOP_FLAG:          DBIT  1  ; 1 = LFDNOOP command received
HKP_RESUME:        DBIT  1  ; 1 = resume prior housekeeping script command
CMD_BEGUN:         DBIT  1  ; 1 = Boot command was started
EXE_FLAG:          DBIT  1  ; 1 = Boot command fully executed
WDOG_ENABLED:      DBIT  1  ; 1 = Watchdog enabled
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;                             Boot-to-Operate ICD
;*****************************************************************************
;*****************************************************************************
   ; BIT variable "IVEC" determines whether interrupts use the BOOT space
   ; or OPERATE space.
   ; If in BOOT (IVEC=0), ISRs jump to the local routines (in PROM).
   ; If in OPERATE (IVEC=1), ISRs jump to offsets 8003h, 800Bh, etc.
   ; to look there for a jump to the actual OPERATE mode ISR.  This
   ; facilitates the parallel development of BOOT and OPERATE with minimal
   ; interface constraints.
   BSEG  at  07Fh
IVEC:              DBIT  1  ; 0=Boot mode ISRs, 1=Operate mode ISRs

$if INCLUDE_TEST_CODE
   ; When LFDJMPCS ("Jump to Operate") is commanded, the boot code
   ; will jump to "OPERATE_MAIN".  At that location there should be
   ; a jump to the actual Operate code reentry point.
   ; This procedure is followed to give the Operate code writer flexibility
   ; in placement of the reentry code without the writer of Boot code
   ; knowing in advance.
   CSEG  at  8000h
OPERATE_MAIN:
   sjmp  $                  ; Loop forever (not in actual operate code)

   ; Dummy Operate-mode ISRs allow us to test the interface in simulation.
   CSEG  at  8003h
   sjmp  $
   CSEG  at  800Bh
   sjmp  $
   CSEG  at  8013h
   sjmp  $
   CSEG  at  801Bh
   sjmp  $
   CSEG  at  8023h
   sjmp  $
   CSEG  at  802Bh
   sjmp  $
   CSEG  at  8033h
   sjmp  $
$else
OPERATE_MAIN   equ   8000h
$endif
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;           Boot Reset Vector and Interrupt Service Routine Vectors
;*****************************************************************************
;*****************************************************************************
   CSEG  at  0000h              ; Reset code must start at 0000h for 8051

   ; Jump to the reset entry point ...
   ljmp   BOOT_MAIN

   ; ISR jump vectors (for BOOT and OPERATE).

   ; Primary Command-Received ISR ...
   CSEG  at  0003h
boot03:
   jb     IVEC, operate03       ; Operate mode ISR 03
   ajmp   ISR_Pri_Cmd_Received  ; Boot mode ISR 03
operate03:
   ljmp   8003h

   ; Timer0 ISR -- executes at a regular interval for main loop timing ...
   CSEG  at  000Bh
   jb     IVEC, operate0B       ; Operate mode ISR 0B
   ajmp   ISR_Timer0            ; Boot mode ISR 0B
operate0B:
   ljmp   800Bh

   ; Secondary Command-Received ISR ...
   CSEG  at  0013h
   jb     IVEC, operate13       ; Operate mode ISR 13
   ajmp   ISR_Sec_Cmd_Received  ; Boot mode ISR 13
operate13:
   ljmp   8013h

   ; Timer1 ISR -- MUX delays for A/D reading
   CSEG  at  001Bh
   jb     IVEC, operate1B       ; Operate mode ISR 1B
   ajmp   ISR_Timer1            ; Boot mode ISR 1B
operate1B:
   ljmp   801Bh

   ; Serial Port ISR -- Not Used
   CSEG  at  0023h
   jb     IVEC, operate23       ; Operate mode ISR 23
   ajmp   ISR_Serial_Port       ; Boot mode ISR 23
operate23:
   ljmp   8023h

   ; Timer2 ISR -- Not Used
   CSEG  at  002Bh
   jb     IVEC, operate2B       ; Operate mode ISR 2B
   ajmp   ISR_Timer2            ; Boot mode ISR 2B
operate2B:
   ljmp   802Bh

   ; PCA ISR -- Not Used
   CSEG  at  0033h
   jb     IVEC, operate33       ; Operate mode ISR 33
   ajmp   ISR_PCA               ; Boot mode ISR 33
operate33:
   ljmp   8033h
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;                 Boot Mode Interrupt Service Routines (ISRs)
;*****************************************************************************
;*****************************************************************************
; Subroutine -- ISR_Pri_Cmd_Received
; Purpose -- Handles External interrupt 0 which is invoked when a 32-bit
;            command word is present in its buffer and ready for reading by
;            the 8051.  This is the PRIMARY command channel interrupt.
; Inputs -- none
; Outputs -- command data value written to memory
; Timing -- 88 machine cycles (0.132 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/21/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Pri_Cmd_Received:
   ISR_SAVE_STATE
   mov   PSW, #(1 SHL 3)    ; use register bank 1

   ; Setup DPTR and fetch command MSB from registers ...
   mov   DPTR, #RXD_HARDWARE + 300h
   lcall RXD_READ           ; move data to command buffer

   ISR_RESTORE_STATE
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_Sec_Cmd_Received
; Purpose -- Handles External interrupt 1 which is invoked when a 32-bit
;            command word is present in its buffer and ready for reading by
;            the 8051.  This is the SECONDARY command channel interrupt.
; Inputs -- none
; Outputs -- command data value written to memory
; Timing -- 88 machine cycles (0.132 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/21/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Sec_Cmd_Received:
   ISR_SAVE_STATE
   mov   PSW, #(1 SHL 3)    ; use register bank 1

   ; Setup DPTR and fetch command MSB from registers ...
   mov   DPTR, #RXD_HARDWARE + 700h
   lcall RXD_READ           ; move data to command buffer

   ISR_RESTORE_STATE
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_Timer0
; Purpose -- Timer0 interrupts occur at a rate of "ticks_per_sec" times per
;            second.  This ISR decrements a counter which, upon reaching zero,
;            indicates that one second has elapsed.  At that time, a 4-byte
;            seconds counter is incremented.  Timer0 is the base timer for the
;            PCA's watchdog timer (allowing it the full 10-second timeout).
;            Timer0 is setup in the 16-bit counter mode and is reloaded each
;            call to a period of '1/ticks_per_sec' seconds.
; Inputs -- none
; Outputs -- mLFCTIME (4-bytes) incremented each second.
; Timing -- Seconds counter not incremented:  51 machine cycles
;           Seconds counter incremented (lower word only):  87 machine cycles
;           Seconds counter incremented (long word):  103 machine cycles
;           Worst case:  103 machine cycles (0.155 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/13/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Timer0:
ticks_per_sec   equ   50
clock_freq      equ   8000000
latency         equ   14h   ; number of machine cycles before timer update
count_value     equ   (65536 - clock_freq/12/ticks_per_sec + latency)
   ISR_SAVE_STATE
   mov   PSW, #(1 SHL 3)    ; use register bank 1

   mov   TL0, #0            ; prevent rollover while loading timer
   mov   TH0, #HIGH(count_value)
   mov   TL0, #LOW(count_value)

   ; Increment the high-resolution timer ...
   mov   DPTR, #mTICKS
   movx  A, @DPTR
   inc   A
   movx  @DPTR, A
   
   ; Monitor stack pointer growth ...
   clr   C                  ; prepare for SUBB operation
   mov   A, SP              ; look at stack pointer for growth
   dec   A                  ; ACC = SP-1
   subb  A, maxSP           ; ACC = (SP-1) - maxSP
   jc    ISR_Timer0_10      ; jump if SP <= maxSP
   mov   maxSP, SP          ; store new (higher) stack pointer value

ISR_Timer0_10:
   
   ; Check for expiration of the one-second interval ...
   djnz  CTR_TICK, ISR_Timer0_99
   ; reset count-down register ...
   mov   CTR_TICK, #ticks_per_sec
   ; increment seconds counter
   mov   DPTR, #mLFCTIME
   lcall INCREMENT_LWORD
   
   ; roll over on the high-resolution timer ...
   mov   DPTR, #mTICKS
   clr   A
   movx  @DPTR, A
   
   setb  SECS_FLAG          ; flag for 1-second based tasks

ISR_Timer0_99:
   ISR_RESTORE_STATE
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_Timer1
; Purpose -- Timer1 interrupts occur at a rate specified by the A/D conversion
;            code.  Since it is often necessary to delay after setting a
;            MUX address and prior to commanding the A/D to start its
;            conversion, Timer1 is used for this purpose.
;            Timer1 is setup in the 16-bit counter mode, so its maximum
;            interval is 65536/(8e6/12) = 98.3 msec (@ 8-MHz)
; Inputs -- none
; Outputs -- flag bit 'TIMER1_DONE' is set to '1' (TRUE)
;            flag bit 'TIMER1_EDGE' is set to '1' if bit 'TIMER1_DONE' changes
;               from low-to-high and cleared otherwise.
; Timing -- 31 machine cycles (0.047 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/21/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Timer1:
   ISR_SAVE_STATE
   mov   PSW, #(1 SHL 3)    ; use register bank 1

   clr   TIMER1_EDGE        ; assume NO timer-edge is occurring now
   jb    TIMER1_DONE, ISR_Timer1_99  ; jump if assumption correct
   setb  TIMER1_DONE        ; flag used for housekeeping background task
   setb  TIMER1_EDGE        ; Indicates low-to-high on 'TIMER1_DONE'
      
ISR_Timer1_99:
   ISR_RESTORE_STATE
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_Timer2
; Purpose -- Unused
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Timer2:
   REPORT_DIAG  DIAG0031    ; Diagnostic:  Attempt to execute nonexistent ISR
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_Serial_Port
; Purpose -- Unused
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_Serial_Port:
   REPORT_DIAG  DIAG0031    ; Diagnostic:  Attempt to execute nonexistent ISR
   reti
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- ISR_PCA
; Note -- The watchdog timer is implemented using the Programmable Counter
;         Array (PCA) module.  This function does not require an interrupt
;         to function.
; Purpose -- Unused
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ISR_PCA:
   REPORT_DIAG  DIAG0031    ; Diagnostic:  Attempt to execute nonexistent ISR
   reti
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;                       Boot-Mode ISR support routines
;*****************************************************************************
;*****************************************************************************
; Subroutine -- RXD_READ
; Purpose -- Read command from I/O registers.  If command is 80000000h,
;            force a Power-on reset.  Otherwise, if command address is in
;            the range 0040h-045Ah, store the 2 command data bytes in the
;            given address (increased by 2000h for boot mode).
; Note -- This routine is only called by "ISR_Pri_Cmd_Received" and
;         "ISR_Sec_Cmd_Received".
; Inputs -- DPTR set to
;             "RXD_HARDWARE+300h" for Primary command receive channel
;                  (external interrupt 0) or
;             "RXD_HARDWARE+700h" for Secondary command channel receive
;                  (external interrupt 1).
;           I.E., DPTR points to MSB of incoming command for the appropriate
;           command channel.
; Outputs -- System reset if command is 80000000h.  Command data stored in
;            address provided if within proper range.  Flag 'RXD_FLAG' set to
;            '1' if a process command was received (i.e. command with address
;            0440h)
; Timing -- See timing for the ISRs which call this routine.
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
RXD_READ:
   movx  A,@DPTR            ; read command MSB (bits 31..24)
   mov   CMDBUFF+3,A        ; store in internal RAM
   dec   DPH                ; point to next command byte (bits 23..16)
   movx  A,@DPTR            ; read command byte
   mov   CMDBUFF+2,A        ; store bits 23..16
   dec   DPH
   movx  A,@DPTR
   mov   CMDBUFF+1,A        ; store bits 15..8
   dec   DPH
   movx  A,@DPTR            ; read LSB (this also clears hardware BUSY line)
   mov   CMDBUFF+0,A        ; store bits 7..0

   ; If command MS-bit is '1' and any of the other bits 31 bits are
   ; non-zero, command word is invalid ...
   ; If command MS-bit is '1' and ALL other bits are zero, force a power-on
   ; reset ...
   ; Otherwise (MS-bit is '0'), process the command word ...
RXD_READ_00:
   mov   A, CMDBUFF+3          ; command MSB
   jnb   ACC.7, RXD_READ_10    ; MS-bit is '0' --> OK
   anl   A, #7Fh               ; mask off MS-bit
   jnz   RXD_READ_90           ; other bits non-zero --> invalid
   mov   A, CMDBUFF+2          ; check next byte
   jnz   RXD_READ_90           ; non-zero bits --> invalid
   mov   A, CMDBUFF+1          ; check next byte
   jnz   RXD_READ_90           ; non-zero bits --> invalid
   mov   A, CMDBUFF+0          ; check next byte
   jnz   RXD_READ_90           ; non-zero bits --> invalid
   ljmp  DCE_LFDRSTP           ; Cmd = 80000000h --> power-on reset.

RXD_READ_10:    ; Valid dword.  Perform address range checking ...

   ; Valid range for writing to memory is 0040h-045Ah (memory upload and
   ; command packet).
   ; ACC holds MSB of command.  Mask off most-sig 3 bits (since they are
   ; "unused") and check range.
   ; Subtract 0040h from CMDBUFF+3:CMDBUFF+2 to ensure that address is
   ; greater than or equal to 0040h ...
   clr   C                  ; prepare for SUBB operation
   mov   A, CMDBUFF+2       ; lower address byte
   subb  A, #LOW(0040h)
   mov   A, CMDBUFF+3       ; upper address byte
   anl   A, #1Fh            ; mask off most-sig 3 bits (unused bits)
   mov   CMDBUFF+3, A       ; store value to preclude need to mask later
   subb  A, #HIGH(0040h)
   jc    RXD_READ_90        ; address less than 0040h -- too low
   
   ; Subtract 045Bh from CMDBUFF+3:CMDBUFF+2 to ensure that address is
   ; less than or equal to 045Ah.
   ; Note:  Carry flag is already clear.
   mov   A, CMDBUFF+2       ; address LSB
   subb  A, #LOW(045Bh)
   mov   A, CMDBUFF+3       ; address MSB
   subb  A, #HIGH(045Bh)
   jnc   RXD_READ_90        ; address greater than 045Ah -- too high

   ; address is in the range 0040h - 045Ah, so store the least-sig 2 bytes
   ; of the command (i.e., the data bytes) in memory ...
   
   mov   A, CMDBUFF+3       ; Put address in DPTR
   mov   DPH, A             ; MSB
   mov   A, CMDBUFF+2
   mov   DPL, A             ; LSB
   ; Boot mode offsets the actual address upward by 2000h
   orl   DPH, #20h
   ; now store the 2 data bytes ...
   mov   A, CMDBUFF+0       ; data LSB
   movx  @DPTR, A
   inc   DPTR
   mov   A, CMDBUFF+1       ; data MSB
   movx  @DPTR, A

   ; See if a command opcode was received (address = RS17)
   setb  RXD_FLAG           ; ASSUME an command opcode received
   mov   DPTR, #RS17        ; address of command opcode LSB
   movx  A, @DPTR
   jnz   RXD_READ_99        ; opcode was received
   clr   RXD_FLAG           ; opcode NOT received
   sjmp  RXD_READ_99        ; done
   
RXD_READ_90:
   ; Increment the "command word received but not stored" counter ...
   mov   DPTR, #mCWNS
   lcall INCREMENT_WORD
   ; Store the "unstored" command word ...
   mov   DPTR, #mWNS
   mov   A, CMDBUFF+0       ; data LSB
   movx  @DPTR, A
   inc   DPTR
   mov   A, CMDBUFF+1       ; data MSB
   movx  @DPTR, A
   inc   DPTR
   mov   A, CMDBUFF+2       ; address LSB
   movx  @DPTR, A
   inc   DPTR
   mov   A, CMDBUFF+3       ; address MSB
   movx  @DPTR, A

; clean up and exit
RXD_READ_99:
   ret                      ; RXD_READ
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;                          Boot-Mode Main Executive
;*****************************************************************************
;*****************************************************************************
; Subroutine -- BOOT_MAIN
; Purpose -- This is the entry for boot code.  Upon any type of reset, the
;            first instruction is to jump to this routine.  Here various
;            internal and external RAM locations are initialized, timers and
;            interrupts are initialized, and finally, a JUMP (no return) is
;            made to the main executive loop.
; Inputs -- none
; Outputs -- none
; Timing -- 56,577 machine cycles (84.9 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/21/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
BOOT_MAIN:

   ; Initialize the processor ...
   mov   IE, #0             ; disable all interrupts
   mov   IP, #0             ; all interrupts at same proiority level
   mov   TCON, #0           ; stop timers
   mov   SP, #_STACK        ; Reset stack to base level

   ; Disable the Programmable Counter Array (PCA) watchdog timer ...
   lcall DISABLE_DOG

   ; Safe the hardware ...
   lcall SAFE_CONTROLS

   ; Clear Internal and External RAM ...
   lcall INIT_RAM

$if INCLUDE_TEST_CODE
   orl   PCON, #00010000b   ; Force system into Power-On Reset mode to get
                            ; patchable constants copied.
$endif

   ;--------------------------------------------------------------------------
   ; Check for reset type.
   ;    PCON.4 = 0 indicates soft (watchdog or other) reset
   ;    PCON.4 = 1 indicates hard (power on) reset
   mov   A, PCON
   jnb   ACC.4, WATCHDOG_RESET ; watchdog or other soft reset occurred

POWER_ON_RESET:
   anl   PCON, #11101111b   ; clear PCON.4
   REPORT_DIAG  DIAG001B    ; Diagnostic:  Power-On Reset

   ; Copy patchable constants from PROM code space to External RAM ...
   lcall COPY_PATCHABLE
   ljmp  BOOT_MAIN_20       ; start main executive loop

WATCHDOG_RESET:
   REPORT_DIAG  DIAG001C    ; Diagnostic:  Watchdog reset

BOOT_MAIN_20:
   ;--------------------------------------------------------------------------
   
   ;--------------------------------------------------------------------------
   ; Setup interrupts ...
   clr   IVEC               ; ensure Boot mode ISRs are used
   setb  EA                 ; global interrupt enable

   ; Setup standard 8051 Timers ...
   mov   TMOD, #11h         ; Timer0 in 16-bit timer mode
                            ; Timer1 in 16-bit timer mode
   mov   CTR_TICK, #ticks_per_sec ; preset countdown timer for 1-second
   setb  TR0                ; start Timer0
   setb  ET0                ; enable Timer0 interrupt
   setb  TR1                ; start Timer1
   setb  ET1                ; enable Timer1 interrupt

   ; Setup Primary command interface ...
   clr   IT0                ; interrupt is low-level triggered
   setb  EX0                ; enable interrupt
   
   ; Setup Secondary command interface ...
   clr   IT1                ; interrupt is low-level triggered
   setb  EX1                ; enable interrupt
   ;--------------------------------------------------------------------------

   ; Start background CRC calculations ...
   mov   REGION_CRC, #1     ; Start background CRC checking with PROM
   lcall INIT_CRC           ; initialize internal variables for this region
   lcall CLEAR_OLD_CRCS     ; clear all 'previous CRC' values to prevent CRC
                            ; comparison errors (initially)

   ljmp  BOOT_MAIN_EXECUTIVE ; start main executive loop
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- BOOT_MAIN_EXECUTIVE
; Purpose -- Main Executive loop cyclically calls all task subroutines.
; Inputs -- none
; Outputs -- none
; Timing -- forever (until jump-to-operate)
; Testing -- Tested by simulation and working, 9/21/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
BOOT_MAIN_EXECUTIVE:

   mov   A, SP
   xrl   A, #_STACK         ; see if SP = _STACK
   jz    BOOT_MAIN_EXECUTIVE_10 ; jump if OK
   REPORT_DIAG   DIAG0017   ; Diagnostic:  Stack Growth
   mov   SP, #_STACK        ; reset the stack pointer

BOOT_MAIN_EXECUTIVE_10:

   ; Increment the performance monitor ...
   mov   DPTR, #mLFDPERF
   lcall INCREMENT_WORD

   ;--------------------------------------------------------------------------
   ; Process a command packet ...
TASK_PROCESS_COMMAND:
   ; Reasons to skip this task for this loop ...
   jnb   RXD_FLAG, TASK_PROCESS_COMMAND_99 ; no opcode received
   ; No reason to skip, so enter the task ...
   clr   RXD_FLAG            ; clear the flag
   lcall PROCESS_COMMAND     ; process the command
TASK_PROCESS_COMMAND_99:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
   ; Process the 1-second timeout events, if any.
TASK_ONE_SECOND:
   ; Reasons to skip this task for this loop ...
   jb    RXD_FLAG,  TASK_ONE_SECOND_99 ; opcode received
   jnb   SECS_FLAG, TASK_ONE_SECOND_99 ; 1-second timer NOT elapsed
   ; No reason to skip, so enter the task ...
   clr   SECS_FLAG           ; clear the flag
   ; ...
$if INCLUDE_TEST_CODE
   lcall test_commands
$endif
   ; ...
TASK_ONE_SECOND_99:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
   ; If dump packet queued, send it.
TASK_SEND_DUMP_PKT:
   ; Reasons to skip this task for this loop ...
   jb    RXD_FLAG,       TASK_SEND_DUMP_PKT_99 ; opcode received
   jb    SECS_FLAG,      TASK_SEND_DUMP_PKT_99 ; 1-second timer elapsed
   jnb   SEND_DUMP_FLAG, TASK_SEND_DUMP_PKT_99 ; dump packet not queued
   ; No reason to skip, so enter the task ...
   clr   SEND_DUMP_FLAG      ; clear the flag
   lcall SEND_DUMP_PKT       ; send the packet
TASK_SEND_DUMP_PKT_99:
   ;--------------------------------------------------------------------------
   
   ;--------------------------------------------------------------------------
   ; If housekeeping packet queued, send it.
TASK_SEND_HKP_PKT:
   ; Reasons to skip this task for this loop ...
   jb    RXD_FLAG,       TASK_SEND_HKP_PKT_99 ; opcode received
   jb    SECS_FLAG,      TASK_SEND_HKP_PKT_99 ; 1-second timer elapsed
   jb    SEND_DUMP_FLAG, TASK_SEND_HKP_PKT_99 ; dump packet queued
   jnb   SEND_HKP_FLAG,  TASK_SEND_HKP_PKT_99 ; housekeeping packet not queued
   ; No reason to skip, so enter the task ...
   clr   SEND_HKP_FLAG       ; clear the flag
   lcall SEND_HKP_PKT        ; send the packet
TASK_SEND_HKP_PKT_99:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
   ; Build Housekeeping packet ...
TASK_BUILD_HKP:
   ; Reasons to skip this task for this loop ...
   jb    RXD_FLAG,       TASK_BUILD_HKP_99 ; opcode received
   jb    SECS_FLAG,      TASK_BUILD_HKP_99 ; 1-second timer elapsed
   jb    SEND_DUMP_FLAG, TASK_BUILD_HKP_99 ; dump packet queued
   jb    SEND_HKP_FLAG,  TASK_BUILD_HKP_99 ; houskeeping packet queued !!
   jnb   HKP_BUILD_FLAG, TASK_BUILD_HKP_99 ; not flagged to build HKP
   jnb   TIMER1_DONE,    TASK_BUILD_HKP_99 ; MUX delay timer not expired
   ; No reason to skip, so enter the task ...
   ;    NOTE:  This is a BACKGROUND task (other events interrupt it), so it
   ;    will clear appropriate flags when it is complete.
   lcall BUILD_HKP          ; Build housekeeping from the script

   ; When finished building housekeeping, clear command buffer to prepare
   ; for next command to arrive ...
   jb    HKP_BUILD_FLAG, TASK_BUILD_HKP_99 ; STILL flagged to build HKP
   lcall CLEAR_CMD_PKT      ; clear 14-word command buffer
   lcall CLEAR_HKP_PKT      ; clear PARTS of the housekeeping work area
TASK_BUILD_HKP_99:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
   ; Compute background CRC ...
TASK_BG_CRC:
   ; Reasons to skip this task for this loop ...
   jb    RXD_FLAG,       TASK_BG_CRC_99 ; opcode received
   jb    SECS_FLAG,      TASK_BG_CRC_99 ; 1-second timer elapsed
   jb    SEND_DUMP_FLAG, TASK_BG_CRC_99 ; dump packet queued
   jb    SEND_HKP_FLAG,  TASK_BG_CRC_99 ; housekeeping packet queued
   jb    TIMER1_EDGE,    TASK_BG_CRC_99 ; Housekeeping A/D MUX delay expired
   jnb   CRC_FLAG,       TASK_BG_CRC_99 ; CRC background task NOT flagged
   ; No reason to skip, so enter the task ...
   ;    NOTE:  This is a BACKGROUND task (other events interrupt it), so it
   ;    will clear appropriate flags when it is complete.
   lcall BG_CRC              ; background CRC task
TASK_BG_CRC_99:
   ;--------------------------------------------------------------------------

   ; Done with tasks for this loop.  Back to the top.
BOOT_MAIN_EXECUTIVE_99:
   ljmp  BOOT_MAIN_EXECUTIVE ; back to top of the loop
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;                    Main-Executive-Loop Support Routines
;*****************************************************************************
;*****************************************************************************
; Subroutine -- PROCESS_COMMAND
; Purpose -- Process a command packet.  Called when RXD_FLAG=1 in main exec.
; Inputs -- Command packet in RAM (buffer address 'RS17')
; Outputs -- Command checked for correctness, executed if correct.  Counters
;            incremented, memory monitors updated, housekeeping packet script
;            initialized for building housekeeping packet.
; Timing -- 920 machine cycles (1.38 msec @ 8-MHz), includes all subroutine
;           calls from this routine, EXCLUDING the time required by the
;           particular Boot-mode function being processed.
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
PROCESS_COMMAND:
   ; Verify command packet conforms to proper format ...
   lcall CHECK_CMD_PACKET   ; return value:  ACC=0 if command packet OK
   jnz   PROCESS_COMMAND_30 ; bad command packet -- send housekeeping
   
   ; Good command packet received ...
   ; If watchdog enabled, feed the dog ...
   jnb   WDOG_ENABLED, PROCESS_COMMAND_10  ; jump if watchdog NOT enabled
   lcall ENABLE_DOG         ; enable AND feed watchdog

PROCESS_COMMAND_10:
   ; Execute the command ...
   lcall COMMAND_VECTOR

   ; Increment packet counter ...
   mov   DPTR, #mLFCPKTL
   lcall INCREMENT_LWORD

   ; Check for LFDNOOP command ...
   jbc   NOP_FLAG, PROCESS_COMMAND_30  ; jump if command is LFDNOOP

   ; Increment commands-received counter ...
   mov   DPTR, #mLFDCMDR
   lcall INCREMENT_WORD

   ; Check if command was executed ...
   jnb   EXE_FLAG, PROCESS_COMMAND_20  ; jump if command NOT executed
   clr   EXE_FLAG
   clr   CMD_BEGUN                     ; must clear "command begun" as well

   ; Increment commands-executed counter   
   mov   DPTR, #mLFDCMDX
   lcall INCREMENT_WORD
   
   ; record the last command executed ...
   mov   DPTR, #RS17        ; command opcode within command buffer
   movx  A, @DPTR
   mov   DPTR, #mLFDLSTC    ; store in 'mLFDLSTC'
   movx  @DPTR, A

   sjmp  PROCESS_COMMAND_30

PROCESS_COMMAND_20:
   jnb   CMD_BEGUN, PROCESS_COMMAND_30  ; jump if cmd never begun or illegal
   clr   CMD_BEGUN

   ; Report that the command did not complete ...
   REPORT_DIAG  DIAG0013    ; Diagnostic:  Command not completed

PROCESS_COMMAND_30:

   ; Setup for housekeeping packet construction ...
   lcall SETUP_TO_BUILD_HKP

PROCESS_COMMAND_99:
   ret                      ; PROCESS_COMMAND
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CLEAR_HKP_PKT
; Purpose -- Clear parts of the housekeeping packet (after packet is sent).
;            Note that the housekeeping packet is not really cleared, but,
;            rather, those working variables which are copied into the
;            housekeeping packet by the routine 'BUILD_HKP'.
; Inputs -- none
; Outputs -- Specific parts of housekeeping packet cleared (see below)
; Timing -- 75 machine cycles (0.113 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/21/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CLEAR_HKP_PKT:

   ; Clear array of 8 HST errors ...
   mov   DPTR, #mLFDERR     ; first XDATA location to clear
   mov   R6, #HIGH(8)       ; number of bytes to clear in R6:R7
   mov   R7, #LOW(8)
   call  CLEAR_RAM          ; clear the memory

   ; Clear Performance monitor ...
   mov   DPTR, #mLFDPERF
   clr   A
   movx  @DPTR, A
   inc   DPTR
   movx  @DPTR, A

   ; Clear the number of diagnostics stored since last housekeeping sent
   mov   DPTR, #mCURR_NUM
   clr   A
   movx  @DPTR, A

CLEAR_HKP_PKT_99:
   ret                      ; CLEAR_HKP_PKT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CLEAR_CMD_PKT
; Purpose -- Clear the 14 command words in order that commanding errors may
;            be detected and flagged.
; Inputs -- none
; Outputs -- 14 command packet words cleared
; Timing -- 
; Testing -- 
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CLEAR_CMD_PKT:
   mov   DPTR, #RS17        ; first XDATA location to clear
   mov   R6, #HIGH(28)      ; number of bytes to clear in R6:R7
   mov   R7, #LOW(28)
   call  CLEAR_RAM          ; clear the memory
CLEAR_CMD_PKT_99:
   ret                      ; CLEAR_HKP_PKT
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;                             Boot-Mode Commands
;*****************************************************************************
;*****************************************************************************
; Subroutine -- DCE_LFDNOOP (DCE Opcode '80'X)
; Purpose -- Force the DCE to send a housekeeping packet and feed the
;            watchdog timer without incrementing the command counter.
; Inputs -- none
; Outputs -- NOP_FLAG=1
; Timing -- 5 machine cycles (0.008 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDNOOP:
   setb  NOP_FLAG           ; main executive loop looks at this flag
   ret                      ; DCE_LFDNOOP
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDMADDR (DCE Opcode '81'X)
; Purpose -- Set memory monitor address for one of the 8 memory monitors
; Inputs -- 3 command parameters:
;    PARAM1 = memory monitor index  0..7
;    PARAM2 = source address        0..FFFF
;    PARAM3 = memory space          0..2 (0=XDATA, 1=CODE, 2=DATA)
; Parameter Range Checking --
;    (1) PARAM1 is masked modulo 8
;    (2) PARAM2 none
;    (3) PARAM3 is masked module 4 and value 3 is rejected
; Outputs -- Memory monitor parameters changed
; Timing -- 125 machine cycles (0.188 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDMADDR:
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Store source address (PARAM2) in R3:R2 ...
   mov   DPTR, #PARAM2      ; Address LSB
   movx  A, @DPTR
   mov   R2, A              ; store address LSB
   inc   DPTR
   movx  A, @DPTR
   mov   R3, A              ; store address MSB

   ; Store memory specifier (PARAM3) in R4 ...
   mov   DPTR, #PARAM3      ; Memory Space LSB
   movx  A, @DPTR
   mov   R4, A              ; store memory space LSB

   ; Store memory monitor number (PARAM1) in R0 ...
   mov   DPTR, #PARAM1      ; Memory monitor index LSB
   movx  A, @DPTR
   anl   A, #7              ; modulo 8
   inc   ACC                ; increment because of DJNZ instruction below
   mov   R0, A              ; load loop counter with index+1

   ; Create bit mask (in ACC) and find offset into 'mLFDMADD' at which to
   ; store the address ...
   mov   DPTR, #mLFDMADD    ; beginning of memory monitor address table
   mov   A, #1              ; bit mask init to bit 0
   sjmp  DCE_LFDMADDR_20
DCE_LFDMADDR_10:   
   rl    A                  ; move bit left
   inc   DPTR               ; addresses are words
   inc   DPTR               ; so we advance pointer two bytes
DCE_LFDMADDR_20:
   djnz  R0, DCE_LFDMADDR_10

   mov   R5, A              ; save mask in R5
   cpl   A                  ; invert the bits to turn one off
   mov   R1, A              ; save inverted mask in R1
   
   ; Store PARAM2 in memory monitor address table
   mov   A, R2              ; store memory monitor address
   movx  @DPTR, A           ; put LSB of addr in table
   inc   DPTR
   mov   A, R3
   movx  @DPTR, A           ; put MSB of addr in table
   
   ; Use memory specifier (PARAM3) to set/clear bits in 'mLFDMDATA'
   ; and 'mLFDMEXTR'.
   mov   A, R4              ; get memory space select bits
   anl   A, #3              ; modulo 4
   rl    A                  ; multiply by two for jump table
   mov   DPTR, #MM_TBL      ; Use jump table below to select command
   jmp   @A+DPTR
MM_TBL:
   sjmp  MM_XDATA           ; XDATA space (external)
   sjmp  MM_CODE            ; CODE space
   sjmp  MM_DATA            ; DATA space (internal 8051)
   sjmp  DCE_LFDMADDR_99    ; Illegal.  Exit.

MM_XDATA:                   ; External data
   mov   DPTR, #mLFDMDATA   ; get bitmap for data/code
   movx  A, @DPTR
   anl   A, R1              ; CLEAR the appropriate bit using mask
   movx  @DPTR, A           ; save bitmap
   mov   DPTR, #mLFDMEXTR   ; get bitmap for internal/external data
   movx  A, @DPTR
   anl   A, R1              ; CLEAR the appropriate bit using mask
   movx  @DPTR, A           ; save bitmap
   sjmp  DCE_LFDMADDR_98    ; success

MM_CODE:                    ; Code
   mov   DPTR, #mLFDMDATA   ; get bitmap for data/code
   movx  A, @DPTR
   orl   A, R5              ; SET the appropriate bit using mask
   movx  @DPTR, A           ; save bitmap
   sjmp  DCE_LFDMADDR_98    ; success

MM_DATA:                    ; Internal data
   mov   DPTR, #mLFDMDATA   ; get bitmap for data/code
   movx  A, @DPTR
   anl   A, R1              ; CLEAR the appropriate bit using mask
   movx  @DPTR, A           ; save bitmap
   mov   DPTR, #mLFDMEXTR   ; get bitmap for internal/external data
   movx  A, @DPTR
   orl   A, R5              ; SET the appropriate bit using mask
   movx  @DPTR, A           ; save bitmap
    
DCE_LFDMADDR_98:
   setb  EXE_FLAG           ; command executed

DCE_LFDMADDR_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDMADDR
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDCRC (DCE Opcode '82'X)
; Purpose -- Setup a background task to compute the CRC of a specified range
;            of memory.  The CRC is computed by routine 'BG_CRC', the
;            background CRC task handler.  'Region 0' is used for the region
;            specified in this command.
; Inputs -- 3 command parameters:
;    PARAM1 = starting address  0..FFFF
;    PARAM2 = number of bytes   0..FFFF
;    PARAM3 = memory space      0..1 (0=XDATA, 1=CODE)
; Parameter Range Checking --
;    (1) PARAM1 N/A
;    (2) PARAM2 N/A
;    (3) PARAM3 only allowed the values '0' or '1'
; Outputs -- While computing the CRC, housekeeping will show CRC=0000h. Once
;            CRC computation is complete, housekeeping will show the computed
;            value in parameter 'LFMCRC'.
; Timing -- 168 machine cycles (0.252 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/13/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDCRC:
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; This command merely populates element 0 of the background CRC table.
   ; The background CRC task is then initiated as for other regions.
   
   ; Read elements from the PARAM list, push them onto the 8051 stack,
   ; then write them to the CRC table ...

   ; Read the source specifier (PARAM3), perform range checking, and if OK
   ; push onto 8051 stack ...
   mov   DPTR, #PARAM3      ; memory space specifier
   movx  A, @DPTR           ; PARAM3 LSB
   mov   R2, A              ; store PARAM3 LSB
   inc   DPTR
   movx  A, @DPTR           ; PARAM3 MSB
   jnz   DCE_LFDCRC_99      ; parameter error -- exit
   mov   R0, #0             ; assume XDATA region
   mov   A, R2
   jz    DCE_LFDCRC_10      ; XDATA is correct region
   mov   R0, #1             ; assume CODE region
   dec   A
   jnz   DCE_LFDCRC_99      ; parameter error -- exit

DCE_LFDCRC_10:
   mov   A, R0
   push  ACC                ; push memory specifier on stack

   ; Read the length (PARAM2) and push onto 8051 stack ...
   mov   DPTR, #PARAM2      ; point to LSB of length
   movx  A, @DPTR
   push  ACC                ; push length LSB onto stack
   inc   DPTR
   movx  A, @DPTR
   push  ACC                ; push length MSB onto stack

   ; Read the starting address (PARAM1) and push onto 8051 stack ...
   mov   DPTR, #PARAM1      ; point to LSB of starting address
   movx  A, @DPTR
   push  ACC                ; push address LSB onto stack
   inc   DPTR
   movx  A, @DPTR
   push  ACC                ; push address MSB onto stack

   ; Read items from stack and put into the background CRC table ...
   mov   DPTR, #CRC_REGIONS ; base address of CRC regions (Region 0)
   pop   ACC                ; pop address MSB from stack
   movx  @DPTR, A           ; store address MSB
   inc   DPTR
   pop   ACC                ; pop address LSB from stack
   movx  @DPTR, A           ; store address LSB
   inc   DPTR
   pop   ACC                ; pop length MSB from stack
   movx  @DPTR, A           ; store length MSB
   inc   DPTR
   pop   ACC                ; pop length LSB from stack
   movx  @DPTR, A           ; store length LSB
   inc   DPTR
   pop   ACC                ; pop memory specifier from stack
   movx  @DPTR, A           ; store memory space specifier
   
   ; Clear 'current CRC' to show 0000h while computing this CRC ...
   mov   DPTR, #mLFMCRC
   clr   A
   movx  @DPTR, A
   inc   DPTR
   movx  @DPTR, A
   ; Clear 'previous CRC' to prevent CRC mis-compare when completed ...
   inc   DPTR
   movx  @DPTR, A
   inc   DPTR
   movx  @DPTR, A

   mov   REGION_CRC, #0     ; LFDCRC uses region '0'
   lcall INIT_CRC           ; initialize internal variables for the new region

DCE_LFDCRC_98:
   setb  EXE_FLAG           ; command executed

DCE_LFDCRC_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDCRC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDCOPY (DCE Opcode '83'X)
; Purpose -- Copies a range of memory (CODE or XDATA) to XDATA.
; Input -- 4 command parameters:
;    PARAM1 = source address       0..FFFF
;    PARAM2 = destination address  0..FFFF
;    PARAM3 = number of bytes      0..FFFF (address rollover is OK)
;    PARAM4 = source memory space  0..1 (0=XDATA, 1=CODE)
;          If the number of bytes is 0000h, then the full 65536 bytes of
;          memory are copied.
; Parameter Range Checking --
;    (1) PARAM1 N/A
;    (2) PARAM2 N/A
;    (3) PARAM3 N/A
;    (4) PARAM4 only allowed the values '0' or '1'
; Output -- XDATA space has bytes copied.
; Timing -- Test case:  N=1024 bytes of CODE space ...
;              Timing = 29,722 machine cycles (44.6 msec @ 8-MHz)
;           Test case:  N=1024 bytes of XDATA space ...
;              Timing = 30,825 machine cycles (46.2 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDCOPY:
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Store source address (PARAM1) in R3:R2
   mov   DPTR, #PARAM1      ; point to source address LSB
   movx  A, @DPTR
   mov   R2, A              ; store source LSB
   inc   DPTR
   movx  A, @DPTR
   mov   R3, A              ; store source MSB
      
   ; Store destination address (PARAM2) in R5:R4
   mov   DPTR, #PARAM2      ; point to destination address LSB
   movx  A, @DPTR
   mov   R4, A              ; store destination LSB
   inc   DPTR
   movx  A, @DPTR
   mov   R5, A              ; store destination MSB
      
   ; Store number of bytes (PARAM3) in R0:R7
   mov   DPTR, #PARAM3      ; point to LSB
   movx  A, @DPTR
   mov   R7, A              ; store LSB
   inc   DPTR
   movx  A, @DPTR
   mov   R0, A              ; store MSB
   
   ; Range check source specifier (PARAM4) and use it to setup 'SRC_FLAG'
   mov   DPTR, #PARAM4      ; point to LSB (only byte that counts)
   movx  A, @DPTR
   mov   R1, A              ; store LSB
   inc   DPTR
   movx  A, @DPTR
   jnz   DCE_LFDCOPY_99     ; parameter error -- exit
   clr   SRC_FLAG           ; assume source is XDATA
   mov   A, R1
   jz    DCE_LFDCOPY_30     ; start copy if source is XDATA space
   setb  SRC_FLAG           ; assume source is CODE
   dec   A
   jnz   DCE_LFDCOPY_99     ; parameter error -- exit

DCE_LFDCOPY_30:             ; copy data
   lcall MEM_COPY

DCE_LFDCOPY_98:
   setb  EXE_FLAG           ; command executed

DCE_LFDCOPY_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDCOPY
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDUPLOD (DCE Opcode 'AD'X)
; Purpose -- Copy an upload packet to the desired XDATA locations if the CRC
;            matches that specified in the parameter list.  If the specified
;            CRC = 0000h, then no CRC comparison is made prior to the memory
;            copy (and the CRC is simply not calculated, to save time).
; Inputs -- 3 command parameters:
;    PARAM1 = destination address  0..FFFF
;    PARAM2 = number of bytes      1..1024 (decimal)
;    PARAM3 = predicted CRC        0..FFFF (0=skip CRC checking)
; Parameter Range Checking --
;    (1) PARAM1 N/A
;    (2) PARAM2 only allowed values 0001h through 0400h
;    (3) PARAM3 N/A
; Outputs -- Specified XDATA locations overwritten.
; Timing -- Test case:  N=1024 bytes
;           w/o CRC --> Timing = 30,832 machine cycles (46.2 msec @ 8-MHz)
;           w/  CRC --> Timing = 70,879 machine cycles (106.3 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDUPLOD:
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2
   
   ; PARAM3 tells us whether we are to calculate the CRC for this upload
   setb  CALC_CRC           ; assume we ARE to calculate the CRC
   mov   DPTR, #PARAM3      ; point to LSB
   movx  A, @DPTR
   jnz   DCE_LFDUPLOD_00    ; jump on non-zero (assumption correct)
   inc   DPTR
   movx  A, @DPTR
   jnz   DCE_LFDUPLOD_00    ; jump on non-zero (assumption correct)
   clr   CALC_CRC           ; assumption wrong -- we are NOT computing CRC

DCE_LFDUPLOD_00:

   ; Store length (PARAM2) in R3:R1 and range check ...
   mov   DPTR, #PARAM2      ; point to length parameter LSB
   movx  A, @DPTR           ; read length LSB
   mov   R1, A              ; store LSB in R1
   inc   DPTR               ; point to length parameter MSB
   movx  A, @DPTR           ; read length MSB
   mov   R3, A              ; store MSB in R3
   ; decrement length and ensure it is in the range 0..3FF (hex)
   ; (but do not store the decremented length)
   clr   C                  ; prepare for SUBB instruction
   mov   A, R1              ; LSB --> ACC
   subb  A, #1              ; subtract 1
   mov   A, R3              ; MSB --> ACC
   subb  A, #0              ; propagate carry (borrow)
   ; Now length should be in the range 0..3FF (hex), so if the MSB exceeds
   ; 3 then the length is too long or too short.  Subtract 4 from the MSB
   ; and if the MSB was >= 4, there is NO CARRY BIT.  Therefore, if the
   ; carry bit is set, we are OK.
   clr   C                  ; prepare for SUBB instruction
   subb  A, #4
   jc    DCE_LFDUPLOD_10    ; jump if length is OK (carry bit set)
   
   REPORT_DIAG  DIAG0001    ; Diagnostic:  Bad upload length
   sjmp  DCE_LFDUPLOD_99    ; exit
   
DCE_LFDUPLOD_10:

   ; temporarily store length parameter (currently in R3:R1) in R0:R7 ...
   mov   A, R3
   mov   R0, A
   mov   A, R1
   mov   R7, A

   jnb   CALC_CRC, DCE_LFDUPLOD_50  ; jump if NOT computing CRC

DCE_LFDUPLOD_20:            ; Compute CRC and compare to specified value
   mov   CRCH, #0FFh        ; initialize CRC to 0FFFFh
   mov   CRCL, #0FFh
   mov   DPTR, #RS01        ; starting address of upload packet

   ; Fixup "number of bytes" if LSB is non-zero (to make DJNZ commands work)
   cjne  R1, #0, DCE_LFDUPLOD_22
   sjmp  DCE_LFDUPLOD_30    ; MSB is OK since LSB=0
DCE_LFDUPLOD_22:
   inc   R3                 ; if LSB<>0 we must increment MSB for looping

DCE_LFDUPLOD_30:            ; read data byte-by-byte and compute CRC
   movx  A, @DPTR
   lcall DOCRC              ; compute CRC for this byte
   inc   DPTR
   djnz  R1, DCE_LFDUPLOD_30 ; loop by decrementing counter LSB
   djnz  R3, DCE_LFDUPLOD_30 ; loop by decrementing counter MSB

   ; Save upload packet CRC (little-endian) ...
   mov   DPTR, #mLFMXFER
   mov   A, CRCL
   movx  @DPTR, A           ; store LSB
   inc   DPTR
   mov   A, CRCH
   movx  @DPTR, A           ; store MSB
   
   ; Compare computed CRC with value provided (PARAM3) ...
   mov   DPTR, #PARAM3      ; point to LSB
   movx  A, @DPTR           ; read CRC parameter LSB
   cjne  A, CRCL, DCE_LFDUPLOD_40  ; jump on miscompare
   inc   DPTR
   movx  A, @DPTR           ; read CRC parameter MSB
   cjne  A, CRCH, DCE_LFDUPLOD_40  ; jump on miscompare
   sjmp  DCE_LFDUPLOD_50    ; CRC compares OK, so proceed with copy ...

DCE_LFDUPLOD_40:
   REPORT_DIAG  DIAG0002    ; Diagnostic:  Bad upload CRC
   sjmp  DCE_LFDUPLOD_99    ; exit
                            
DCE_LFDUPLOD_50:            ; Copy upload data to XDATA space
   ; setup registers for MEM_COPY routine ...
   ; R0:R7 already holds "number of bytes" to copy
   ; R3:R2 must hold source address ...
   mov   R3, #HIGH(RS01)    ; address of upload packet
   mov   R2, #LOW(RS01)
   ; R5:R4 must hold destination address (PARAM1) ...
   mov   DPTR, #PARAM1      ; point to LSB of destination address
   movx  A, @DPTR
   mov   R4, A
   inc   DPTR               ; point to MSB of destination address
   movx  A, @DPTR
   mov   R5, A
   clr   SRC_FLAG           ; source memory specifier (0=XDATA, 1=CODE)
   lcall MEM_COPY           ; copy the data

$if GEOFF_CODE_001004
   lcall CLEAR_OLD_CRCS     ; clear all 'previous CRC' values to prevent CRC
                            ; comparison errors due to this upload
$endif

DCE_LFDUPLOD_98:
   setb  EXE_FLAG           ; command executed
   
DCE_LFDUPLOD_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDUPLOD
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDDNLOD (DCE Opcode 'AE'X)
; Purpose -- Copy a range of XDATA into the download packet space.  Compute a
;            CRC on the specified range of memory and report it in the
;            housekeeping packet.
; Inputs -- 2 command parameters:
;    PARAM1 = memory address   0..FFFF
;    PARAM2 = number of bytes  1..1024 (decimal)
; Parameter Range Checking --
;    (1) PARAM1 N/A
;    (2) PARAM2 only allowed values 0001h through 0400h
; Outputs -- Download packet of specified XDATA locations.
; Timing -- Test case:  N=1024 bytes
;              Timing = 70,869 machine cycles (106.3 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDDNLOD:
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2
   
   ; Store length (PARAM2) in R3:R1 and range check ...
   mov   DPTR, #PARAM2      ; point to length parameter LSB
   movx  A, @DPTR           ; read length LSB
   mov   R1, A              ; store LSB in R1
   inc   DPTR               ; point to length parameter MSB
   movx  A, @DPTR           ; read length MSB
   mov   R3, A              ; store MSB in R3
   ; decrement length and ensure it is in the range 0..3FF (hex)
   ; (but do not store the decremented length)
   clr   C                  ; prepare for SUBB instruction
   mov   A, R1              ; LSB --> ACC
   subb  A, #1              ; Subtract '1'
   mov   A, R3              ; MSB --> ACC
   subb  A, #0              ; propagate carry (borrow)
   ; Now length should be in the range 0..3FF (hex), so if the MSB exceeds
   ; 3 then the length is too long or too short.  Subtract 4 from the MSB
   ; and if the MSB was >= 4, there is NO CARRY BIT.  Therefore, if the
   ; carry bit is set, we are OK.
   clr   C                  ; prepare for SUBB instruction
   subb  A, #4
   jc    DCE_LFDDNLOD_10    ; jump if length is OK (carry bit set)
   
   REPORT_DIAG  DIAG0003    ; Diagnostic:  Bad download length
   sjmp  DCE_LFDDNLOD_99    ; exit

DCE_LFDDNLOD_10:

   ; temporarily store length parameter (currently in R3:R1) in R0:R7 ...
   mov   A, R3
   mov   R0, A
   mov   A, R1
   mov   R7, A

DCE_LFDDNLOD_20:            ; Compute CRC and compare to specified value
   mov   CRCH, #0FFh        ; initialize the CRC to 0FFFFh
   mov   CRCL, #0FFh
   
   ; Put starting address of download block (PARAM1) in DPTR ...
   mov   DPTR, #PARAM1      ; point to LSB of source address
   movx  A, @DPTR
   mov   R4, A
   inc   DPTR               ; point to MSB of source address
   movx  A, @DPTR
   mov   DPH, A
   mov   DPL, R4
   
   ; store this starting address temporarily (for MEM_COPY to follow) ...
   push  DPH
   push  DPL

   ; Fixup "number of bytes" if LSB is non-zero (to make DJNZ commands work)
   cjne  R1, #0, DCE_LFDDNLOD_22
   sjmp  DCE_LFDDNLOD_30    ; MSB is OK since LSB=0
DCE_LFDDNLOD_22:
   inc   R3                 ; if LSB<>0 we must increment MSB for looping

DCE_LFDDNLOD_30:            ; read data byte-by-byte and compute CRC
   movx  A, @DPTR
   lcall DOCRC              ; compute CRC for this byte
   inc   DPTR
   djnz  R1, DCE_LFDDNLOD_30 ; loop by decrementing counter LSB
   djnz  R3, DCE_LFDDNLOD_30 ; loop by decrementing counter MSB

   ; Save download packet CRC (little-endian) ...
   mov   DPTR, #mLFMXFER
   mov   A, CRCL
   movx  @DPTR, A
   inc   DPTR
   mov   A, CRCH
   movx  @DPTR, A

   ; flag main routine to send dump packet ...
   setb  SEND_DUMP_FLAG
   
   ; recall DPTR from stack (for the MEM_COPY operation)
   pop   DPL
   pop   DPH

DCE_LFDDNLOD_50:            ; Copy XDATA to the download space
   ; setup registers for MEM_COPY routine ...
   ; R0:R7 already holds "number of bytes" to copy
   ; R3:R2 must hold source address ...
   mov   R3, DPH            ; address of download packet
   mov   R2, DPL
   ; R5:R4 must hold destination address ...
   mov   R5, #HIGH(TS01)    ; address of download area
   mov   R4, #LOW(TS01)
   clr   SRC_FLAG           ; source memory specifier (0=XDATA, 1=CODE)
   lcall MEM_COPY           ; copy the data

DCE_LFDDNLOD_98:
   setb  EXE_FLAG           ; command executed
   
DCE_LFDDNLOD_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDDNLOD
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDGOTO (DCE Opcode 'EA'X)
; Purpose -- Transfer 8051 control to the address specified by PARAM1.
;            This is a dangerous command!
; Inputs -- 1 command parameter:
;    PARAM1 = jump-to address   0..FFFF (hex)
; Parameter Range Checking --
;    (1) PARAM1 N/A
; Outputs -- Jump to the specified address (must RETURN from there)
; Timing -- 19 machine cycles (0.029 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDGOTO:
   ; Put jump-to address (PARAM1) in DPTR ...
   mov   DPTR, #PARAM1      ; point to LSB of jump address
   movx  A, @DPTR
   push  ACC                ; store LSB
   inc   DPTR               ; point to MSB of jump address
   movx  A, @DPTR
   mov   DPH, A             ; DPH <-- MSB
   pop   DPL                ; DPL <-- LSB
   clr   A                  ; prepare for JMP command

DCE_LFDGOTO_98:
   setb  EXE_FLAG           ; flag command executed
   
DCE_LFDGOTO_99:
   jmp   @A+DPTR            ; DCE_LFDGOTO
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDWDOG (DCE Opcode 'F1'X)
; Purpose -- Enables or Disables the Watchdog Timer.
; Inputs -- 1 command parameter:
;    PARAM1 = disable/enable watchdog  0..1  (0=disable, 1=enable)
; Parameter Range Checking --
;    (1) PARAM1 only allowed the values '0' or '1'
; Outputs -- Watchdog timer disable/enabled, as desired.
; Timing -- 57 machine cycles to Enable, 42 to Disable ( < 0.086 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDWDOG:
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Range check PARAM1 ...
   mov   DPTR, #PARAM1      ; point to LSB of PARAM1
   movx  A, @DPTR
   mov   R0, A              ; store LSB
   inc   DPTR               ; point to MSB
   movx  A, @DPTR
   jnz   DCE_LFDWDOG_99     ; parameter error -- exit
   mov   A, R0              ; LSB of parameter 1
   jz    DCE_LFDWDOG_20     ; disable watchdog
   dec   A                  ; decrement (to look for '1')
   jz    DCE_LFDWDOG_30     ; enable watchdog
   sjmp  DCE_LFDWDOG_99     ; parameter error -- exit
   
DCE_LFDWDOG_20:
   lcall DISABLE_DOG
   sjmp  DCE_LFDWDOG_98
   
DCE_LFDWDOG_30:
   lcall ENABLE_DOG
   
DCE_LFDWDOG_98:
   setb  EXE_FLAG           ; command executed
   
DCE_LFDWDOG_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDWDOG
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDJMPCS (DCE Opcode 'F3'X)
; Purpose -- Jumps to the re-entry point for OPERATE mode (i.e., 8000h).  IAW
;            the ICD, the Operate code will have placed a jump to the actual
;            entry point at address 8000h.
; Inputs -- 1 command parameter:
;    PARAM1 = code space to which to jump  0..2  (0 --> no jump, 1,2 --> Jump)
; Parameter Range Checking --
;    (1) PARAM1 only allowed the values '0', '1', or '2'
; Outputs -- none
; Timing -- 37 machine cycles (0.056 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDJMPCS:
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2
   
   mov   DPTR, #PARAM1
   movx  A, @DPTR
   mov   R0, A              ; save parameter LSB
   inc   DPTR               ; point to parameter MSB
   movx  A, @DPTR
   jnz   DCE_LFDJMPCS_99    ; Error:  MSB<>0
   mov   A, R0              ; parameter LSB
   jz    DCE_LFDJMPCS_99    ; Parameter = 0 --> don't jump to operate
   dec   A
   jz    DCE_LFDJMPCS_10    ; Parameter = 1 --> Jump to Operate
   dec   A
   jz    DCE_LFDJMPCS_10    ; Parameter = 2 --> Jump to Operate
   sjmp  DCE_LFDJMPCS_99    ; Error:  Parameter > 2

DCE_LFDJMPCS_10:
   mov   IE, #0             ; disable all interrupts
   setb  IVEC               ; ensure Operate mode ISRs are used
   mov   PSW, #0            ; use register bank 0 in Operate
   mov   SP, #_STACK        ; Reset stack to base level
   mov   TCON, #0           ; clear timer settings (i.e. stop timers)
   ljmp  OPERATE_MAIN       ; DCE_LFDJMPCS

   ; Parameter error or parameter=0 --> Don't Jump to Operate
DCE_LFDJMPCS_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDJMPCS
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDDIAGC (DCE Opcode 'F4'X)
; Purpose -- Clear the 32 entries of the diagnostic push-down stack.  It DOES
;            NOT clear the incrementing diagnostic serial number reported with
;            the diagnostics in housekeeping.  Nor does this routine affect
;            HST status buffer messages reported separately in housekeeping.
; Inputs -- none
; Outputs -- Diagnostic stack cleared.
; Timing -- 399 machine cycles (0.599 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDDIAGC:
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2
   
   mov   R0, #64            ; 32 2-byte entries to clear
   mov   DPTR, #mLFDDIAGS   ; address of diagnostic push-down stack
   clr   A                  ; fill value
DCE_LFDDIAGC_00:
   movx  @DPTR, A           ; zero-out location
   inc   DPTR               ; advance pointer to next location
   djnz  R0, DCE_LFDDIAGC_00 ; loop until done

DCE_LFDDIAGC_98:
   setb  EXE_FLAG           ; command executed

   pop   PSW                ; restore register bank to previous
   ret                      ; DCE_LFDDIAGC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDRSTP (DCE Opcode 'F0'X) and
;               DCE_LFDRSTW (DCE Opcode 'F5'X)
; Purpose -- Initiate a reset by commanding.
; Inputs -- none
; Outputs -- System reset.  PCON.4 will be set ('DCE_LFDRSTP') or left cleared
;            ('DCE_LFDRSTW'), as appropriate.
; Timing -- First toggle of 'HARDRESET' line within 5 machine cycles.
;           Succeeding toggles are 5 msec apart (@ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00.  Cannot produce the
;            actual reset action in simulation, but port pin toggled properly.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDRSTP:
   orl   PCON, #10h         ; setting PCON.4 indicates Power-on-reset
    
DCE_LFDRSTW:
   cpl   HARDRESET          ; toggle reset line to trip POR circuit
    
   ; The Reset doesn't happen until we complement the HARDRESET bit again.
   ; We will toggle the HARDRESET line up to 64 more times before reporting
   ; as a diagnostic the failure to reset.  Each toggle is timed for 5 ms
   ; based on an 8 MHz oscillator (10 msec period).
   
   ; The reason the loop count was changed from 256 to 64 is that 256 loops
   ; of 5 msec yields 1.28 seconds.  If the reset line is, in fact, damaged,
   ; we will find ourselves processing the return from this failed call
   ; while another command packet has already arrived.

reset_delay   equ   1914    ; yields 5 msec toggle period

   mov   R2, #64            ; toggle 64 more times before reporting diag
DCE_LFDRSTW_00:
   mov   R0, #LOW(reset_delay)  ; delay counter LSB
   mov   R1, #HIGH(reset_delay) ; delay counter MSB
DCE_LFDRSTW_10:
   djnz  R0, DCE_LFDRSTW_10 ; decrement counter LSB
   djnz  R1, DCE_LFDRSTW_10 ; decrement counter MSB
   cpl   HARDRESET          ; toggle reset line for HARDRESET
   djnz  R2, DCE_LFDRSTW_00 ; decrement the "64-toggles" counter

   ; We could get here if the reset line is not connected, as in breadboard.
   REPORT_DIAG  DIAG001F    ; Diagnostic:  Commanded reset failure
   ret                      ; DCE_LFDRSTP & DCE_LFDRSTW
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DCE_LFDHKREQ (DCE Opcode 'FF'X)
; Purpose -- Causes COS DCE FSW to send a housekeeping packet (and increment
;            the commands-received counter).
; Inputs -- none
; Outputs -- none
; Timing -- 5 machine cycles (0.008 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DCE_LFDHKREQ:
   setb  EXE_FLAG           ; command executed
   ret                      ; DCE_LFDHKREQ
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CMD_ILLEGAL
; Purpose -- Issues diagnostic for an illegal command.
; Inputs -- none
; Outputs -- Diagnostic message
; Timing -- 909 machine cycles (1.36 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CMD_ILLEGAL:
   clr   CMD_BEGUN          ; don't flag illegal commands as begun
   REPORT_DIAG  DIAG0011    ; Diagnostic:  Illegal Command
   ret                      ; CMD_ILLEGAL
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CMD_OPERATE
; Purpose -- Issues diagnostic for an Operate-mode command requested in Boot.
; Inputs -- none
; Outputs -- Diagnostic message
; Timing -- 909 machine cycles (1.36 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CMD_OPERATE:
   clr   CMD_BEGUN          ; don't flag Operate mode commands as begun
   REPORT_DIAG  DIAG002F    ; Diagnostic:  Operate mode only
   ret                      ; CMD_OPERATE
;-----------------------------------------------------------------------------


;*****************************************************************************
;*****************************************************************************
;                              Support Routines
;*****************************************************************************
;*****************************************************************************
; Subroutine -- COMMAND_VECTOR
; Purpose -- Uses the lookup table 'COS_BOOT_TBL' to find the address of the
;            boot mode command function specified and jumps to that routine.
; Inputs -- none
; Outputs -- JUMPS to the specified subroutine.  Returning from that routine
;            sends the code back to the place where this function was called
;            (subroutine 'PROCESS_COMMAND').
; Timing -- 44 machine cycles (0.066 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
COMMAND_VECTOR:
   setb  CMD_BEGUN          ; 1=command format OK --> we WILL jump to it
   
   mov   DPTR, #RS17        ; point to Command opcode in receive buffer
   movx  A, @DPTR           ; load opcode
   anl   A, #7Fh            ; mask off msb
   mov   DPTR, #COS_BOOT_TBL ; Point to Command jump table
   lcall CODE_TLU_WORD      ; 2-byte look-up value returned in B:A
   mov   DPH, B             ; put B:A into DPTR for code jump to follow
   mov   DPL, A
   clr   A                  ; prepare for code jump
   jmp   @A+DPTR            ; COMMAND_VECTOR
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- INCREMENT_WORD
; Purpose -- Increment a word pointed to by DPTR.  Assumed little-endian.
; Inputs -- DPTR to word to be incremented as little-endian value
; Outputs -- DPTR incremented twice; ACC changed.
; Timing -- 18 machine cycles (0.027 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
INCREMENT_WORD:
   movx  A, @DPTR           ; fetch LSB
   add   A, #1              ; add 1
   movx  @DPTR, A           ; store LSB
   inc   DPTR               ; point to MSB
   movx  A, @DPTR           ; fetch MSB
   addc  A, #0              ; propagate carry bit
   movx  @DPTR, A           ; store MSB
   inc   DPTR               ; point to byte beyond
   ret                      ; INCREMENT_WORD
;-----------------------------------------------------------------------------   

;*****************************************************************************
; Subroutine -- INCREMENT_LWORD
; Purpose -- Increment a long word pointed to by DPTR.  Assumed little-endian.
; Inputs -- DPTR to word to be incremented as little-endian value
; Outputs -- DPTR incremented 2 or 4 times, depending on carry; ACC changed.
; Timing -- 42 machine cycles (0.063 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
INCREMENT_LWORD:
   lcall INCREMENT_WORD     ; increment lower 2 bytes
   jnc   INCREMENT_LWORD_99 ; if no carry, we're done
   lcall INCREMENT_WORD     ; increment upper 2 bytes
INCREMENT_LWORD_99:
   ret                      ; INCREMENT_LWORD
;-----------------------------------------------------------------------------   

;*****************************************************************************
; Subroutine -- MEM_COPY
; Purpose -- Copies a range of memory (CODE or XDATA) to XDATA.
; Inputs -- source address =      R3:R2
;           destination address = R5:R4
;           number of bytes =     R0:R7
;           source memory space = SRC_FLAG (0=XDATA, 1=CODE)
; Outputs -- XDATA memory changed
; Timing -- Test case:  N=1024 bytes
;              Timing = 30,859 machine cycles (46.3 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
MEM_COPY:

   ; fixup number of bytes if LSB is non-zero (to make DJNZ commands work)
   cjne  R7, #0, MEM_COPY_10
   sjmp  MEM_COPY_30
MEM_COPY_10:
   inc   R0                 ; if LSB<>0 we must increment MSB for looping

MEM_COPY_30:                ; copy data
   mov   DPH, R3            ; get source pointer
   mov   DPL, R2
   jb    SRC_FLAG, MEM_COPY_34  ; jump if source is CODE space
   ; source is XDATA
   movx  A, @DPTR           ; read byte from XDATA
   sjmp  MEM_COPY_36

MEM_COPY_34:                ; copy from CODE to XDATA
   clr   A                  ; prepare for MOVC instruction
   movc  A, @A+DPTR         ; read byte from CODE

MEM_COPY_36:                ; write the data
   inc   DPTR               ; advance source pointer and re-save it
   mov   R3, DPH
   mov   R2, DPL
   mov   DPH, R5            ; get destination pointer
   mov   DPL, R4
   movx  @DPTR, A           ; write byte to XDATA
   inc   DPTR               ; advance destination pointer and re-save it
   mov   R5, DPH
   mov   R4, DPL
   djnz  R7, MEM_COPY_30    ; loop by decrementing counter LSB
   djnz  R0, MEM_COPY_30    ; loop by decrementing counter MSB

MEM_COPY_99:
   ret                      ; MEM_COPY
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CHECK_CMD_PACKET
; Purpose -- Verifies proper syntax of a command packet.
; Inputs -- Command packet at XDATA location 'RS17'
; Outputs -- ACC=0 if packet is OK, ACC=0FFh if error(s)
; Timing -- 223 machine cycles (0.335 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CHECK_CMD_PACKET:

   mov   DPTR, #RS17        ; point to first byte of command packet
   movx  A, @DPTR           ; load opcode
   jb    ACC.7, CHECK_CMD_PACKET_05 ; Command MS-bit must be '1'

   REPORT_DIAG  DIAG0004     ; Diagnostic:  Command MS-bit must be '1'
   mov   A, #0FFh            ; indicate error
   sjmp  CHECK_CMD_PACKET_99 ; done

CHECK_CMD_PACKET_05:

   ; Opcode structure is special.  High and low bytes must be the same.
   movx  A, @DPTR           ; Opcode low byte
   mov   CMDBUFF+0, A
   inc   DPTR
   movx  A, @DPTR           ; Opcode high byte
   cjne  A, CMDBUFF+0, CHECK_CMD_PACKET_10   ; bad packet structure

   ; Opcode structure is OK, so check for command complements
   mov   R0, #7              ; check 7 pairs of words for complements
   mov   DPTR, #RS17         ; reset DPTR pointer
   sjmp  CHECK_CMD_PACKET_20

CHECK_CMD_PACKET_10:
   REPORT_DIAG  DIAG0005     ; Diagnostic:  Command byte must have twin
   mov   A, #0FFh            ; indicate error
   sjmp  CHECK_CMD_PACKET_99 ; done

CHECK_CMD_PACKET_20:
   movx  A, @DPTR           ; parameter LSB
   mov   CMDBUFF+0, A
   inc   DPTR
   movx  A, @DPTR           ; parameter MSB
   mov   CMDBUFF+1, A
   inc   DPTR
   movx  A, @DPTR           ; parameter complement LSB
   mov   CMDBUFF+2, A
   inc   DPTR
   movx  A, @DPTR           ; parameter complement MSB
   inc   DPTR
   cpl   A                  ; complement the parameter complement MSB
   cjne  A, CMDBUFF+1, CHECK_CMD_PACKET_30   ; bad packet structure
   mov   A, CMDBUFF+2       ; parameter complement LSB
   cpl   A                  ; complement the parameter complement LSB
   cjne  A, CMDBUFF+0, CHECK_CMD_PACKET_30   ; bad packet structure
   djnz  R0, CHECK_CMD_PACKET_20
   mov   A, #0              ; indicate no errors
   sjmp  CHECK_CMD_PACKET_99

CHECK_CMD_PACKET_30:
   REPORT_DIAG  DIAG0006    ; Diagnostic:  Command Parameters bad complements
   mov   A, #0FFh           ; indicate error

CHECK_CMD_PACKET_99:
   ret
;-----------------------------------------------------------------------------   

;*****************************************************************************
; Subroutine -- ENABLE_DOG
; Purpose -- Setup and enable the watchdog timer using the Programmable
;    Counter Array (PCA) module.  PCA will use the Timer0 rollover to clock
;    its timer, so Timer0 must be running (the default state in boot).
;    Calling this routine also has the effect of feeding the
;    watchdog timer (i.e., resetting its timeout to the full value).
; Inputs -- PCA registers
; Outputs -- PCA registers
; Timing -- 29 machine cycles (0.044 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
ENABLE_DOG:
watchdog_count  equ   10*ticks_per_sec
   clr   EA                 ; hold off interrupts briefly
   mov   CCON, #0           ; stop the PCA timer
   mov   CL, #0             ; clear PCA timer LSB
   mov   CH, #0             ; clear PCA timer MSB
   mov   CCAP4L, #LOW(watchdog_count)  ; set watchdog timeout LSB
   mov   CCAP4H, #HIGH(watchdog_count) ; set watchdog timeout MSB
   mov   CMOD, #44h         ; Enable WDT mode, use Timer0 rollover
   mov   CCAPM4, #48h       ; PCA module 4 as watchdog
   mov   CCON, #40h         ; start the PCA timer
   setb  WDOG_ENABLED       ; flag bit
   setb  EA                 ; re-enable interrupts
ENABLE_DOG_99:   
   ret                      ; ENABLE_DOG
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DISABLE_DOG
; Purpose -- Disable the watchdog timer using the Programmable Counter Array
;    (PCA) module.  This routine simply stops the PCA timer so that it cannot
;    timeout.
; Inputs -- PCA registers
; Outputs -- PCA registers
; Timing -- 17 machine cycles (0.026 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DISABLE_DOG:
   mov   CMOD, #0           ; Disable WDT mode
   mov   CCAPM4, #0         ; PCA module 4 -- NO watchdog
   mov   CCON, #0           ; stop the PCA timer
   clr   WDOG_ENABLED       ; flag bit
DISABLE_DOG_99:
   ret                      ; DISABLE_DOG
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- SETUP_TO_BUILD_HKP
; Purpose -- Initializes various memory locations (e.g., status bits) prior to
;            building the housekeeping packet.
; Inputs -- none
; Outputs -- none
; Timing -- 
; Testing -- 
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
SETUP_TO_BUILD_HKP:

   ; Status Bits '0' ... Nothing to do for boot.
   ; Status Bits '1' ... Nothing to do for boot.

   ; Status Bits '2' ... Only watchdog enable status to report for boot.
   mov   DPTR, #mLFSBITS2+1 ; MSB of status word 2
   movx  A, @DPTR
   setb  ACC.6              ; assume watchdog ENABLED
   jb    WDOG_ENABLED, SETUP_TO_BUILD_HKP_10
   clr   ACC.6              ; watchdog is DISABLED
SETUP_TO_BUILD_HKP_10:
   movx  @DPTR, A

   ; Status Bits '3' ... Two bilevel bytes.
   mov   DPTR, #HV_STATUS   ; HV_STATUS = 4000h
   movx  A, @DPTR           ; read HV_STATUS
   mov   DPTR, #mLFSBITS3+0 ; Address of LSB of Status Bits '3'
   movx  @DPTR, A           ; write HV_STATUS to Status Bits '3' LSB
   mov   DPTR, #BILEVEL     ; BILEVEL = 4500h
   movx  A, @DPTR           ; read BILEVEL
   mov   DPTR, #mLFSBITS3+1 ; Address of MSB of Status Bits '3'
   movx  @DPTR, A           ; write BILEVEL to Status Bits '3' MSB

   ; Status Bits '4' ... CONTROLS1.
   mov   DPTR, #CONTROLS1   ; CONTROLS1 = 4200h
   movx  A, @DPTR           ; read CONTROLS1 value
   mov   DPTR, #mLFSBITS4   ; Address of Status Bits '4'
   movx  @DPTR, A           ; Status Bits '4' <== CONTROLS1 reading

   ; Status Bits '5' ... CONTROLS2.
   clr   C2_LOCK            ; Unlock CONTROLS2
   mov   DPTR, #CONTROLS2   ; CONTROLS2 = 4300h
   movx  A, @DPTR           ; read CONTROLS2 value
   setb  C2_LOCK            ; Lock CONTROLS2
   mov   DPTR, #mLFSBITS5   ; Address of Status Bits '5'
   movx  @DPTR, A           ; Status Bits '5' <== CONTROLS2 reading

   ; LFDDETID ... Concatenation of LFDDETID1 and LFDDETID0 bits.
   clr   A
   push  ACC
   mov   DPTR, #HV_STATUS   ; HV_STATUS = 4000h
   movx  A, @DPTR
   jnb   ACC.7, SETUP_TO_BUILD_HKP_20
   pop   ACC
   setb  ACC.0
   push  ACC
SETUP_TO_BUILD_HKP_20:
   mov   DPTR, #BILEVEL     ; BILEVEL = 4500h
   movx  A, @DPTR
   jnb   ACC.7, SETUP_TO_BUILD_HKP_21
   pop   ACC
   setb  ACC.1
   push  ACC
SETUP_TO_BUILD_HKP_21:
   pop   ACC
   mov   DPTR, #mLFDDETID
   movx  @DPTR, A

   ; We want the 5 housekeeping mnemonics LFDDIAG0, LFDDIAG1, etc, to show
   ; only the diagnostics stored since the last housekeeping packet was
   ; sent.  If any exist, we'll copy them to the array 'mCURR_5' for the
   ; routine 'BUILD_HKP' to then copy into housekeeping.
   ; Copy the first 'mCURR_NUM' diagnostics from the 32-element push-down
   ; stack onto the list of current 5 (up to 5) diagnostics for housekeeping.
   ; First clear the 5 locations ...
   mov   DPTR, #mCURR_5     ; top of our 'list of 5'
   mov   R6, #HIGH(10)      ; 2 bytes for each = 10 bytes to clear
   mov   R7, #LOW(10)
   lcall CLEAR_RAM
   ; Now find out how many diagnostics were stored during this 1-second
   ; interval ...
   mov   DPTR, #mCURR_NUM   ; number of diagnostics stored
   movx  A, @DPTR
   jz    SETUP_TO_BUILD_HKP_40 ; no diagnostics this 1-second interval
   add   A, #-6             ; subtract 6 (which is one too many)
   jnc   SETUP_TO_BUILD_HKP_30 ; carry indicates > 5 (too many)
   mov   A, #-1             ; prepare to add 6 ==> -1 + 6 = 5
SETUP_TO_BUILD_HKP_30:
   add   A, #6              ; ACC holds number of diagnostics to copy
   rl    A                  ; double value (2 bytes per word)
   ; Setup for call to MEM_COPY ...
   ; Number of BYTES to copy goes in R0:R7 ...
   mov   R7, A
   mov   R0, #0
   ; Source address goes in R3:R2 ...
   mov   R3, #HIGH(mLFDDIAGS)   ; top of push-down diagnostics stack
   mov   R2, #LOW(mLFDDIAGS)
   ; Destination address goes in R5:R4 ...
   mov   R5, #HIGH(mCURR_5)     ; list of 5 current diagnostics
   mov   R4, #LOW(mCURR_5)
   ; Set SRC_FLAG='0' for XDATA copy ...
   clr   SRC_FLAG
   ; make the call ...
   lcall MEM_COPY

SETUP_TO_BUILD_HKP_40:

   ; Refresh the memory monitors ...
   lcall MEMORY_MONITORS

   ; Initialize a pointer to the beginning of the housekeeping memory
   ; (for housekeeping data storage) ...
   mov   HKP_PTR+0, #HIGH(TS17) ; MSB of housekeeping packet start
   mov   HKP_PTR+1, #LOW(TS17)  ; LSB

   ; Initialize a pointer to the beginning of the housekeeping script
   ; which directs the building of the housekeeping packet ...
   mov   SCRIPT_PTR+0, #HIGH(hkp_script)  ; MSB of housekeeping script
   mov   SCRIPT_PTR+1, #LOW(hkp_script)   ; LSB

   ; Indicate that no items are left over from previous runs of the script ...
   mov   HKP_PARAMS+3, #0
   
   setb  HKP_BUILD_FLAG     ; flag housekeeping construction started
   clr   HKP_RESUME         ; flag that we are NOT in an A/D MUX delay
   
SETUP_TO_BUILD_HKP_99:
   ret                      ; SETUP_TO_BUILD_HKP
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- BUILD_HKP
; Purpose -- Builds the housekeeping packet in accordance with the build
;            script (see "patchcst.a51" for the script and "common.inc" for
;            a list of the script opcodes.)
; Inputs -- SCRIPT_PTR+0:SCRIPT_PTR+1 points to the next line of the
;           housekeeping script.
;           HKP_PARAMS+0, +1, +2, & +3 hold the 4 bytes of the current line
;           of the script.  If HKP_PARAMS+3 (the repetition counter) is zero,
;           it is time to read a new line of the script.
;           HKP_PTR+0:HKP_PTR+1 points to the address within the
;           housekeeping buffer at which to write the next piece of data
;           collected by this routine.
;           If a particular line of the housekeeping script has a repetition
;           counter greater than 1 (i.e., write more than 1 byte to
;           housekeeping), the rep. counter is decremented once for each
;           data item collected and the HKP_PTR is incremented once.
;           NOTE:  When preparing to start housekeeping building, making
;           HKP_PARAMS+3 zero will force a reading of the next (i.e. first)
;           line of the housekeeping script.  Of course, both SCRIPT_PTR
;           and HKP_PTR should have been set up prior to this first call.
; Outputs -- The housekeeping buffer is populated in accordance with the
;            opcodes and parameters found in the housekeeping script.
; Timing -- Using the default housekeeping packet ...
;           w/ MUX delay 0 -- 23,591 machine cycles (35.4 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/21/00
;*****************************************************************************
$SET (STAY_IN_BUILD_HKP=1)

FILL_TO_BOUNDARY   16, 0FFh
BUILD_HKP:

$if STAY_IN_BUILD_HKP
BUILD_HKP_NEXT:
   ; Reasons to leave this background task prior to completion ...
   jb    RXD_FLAG,  BUILD_HKP_10  ; opcode received
   jb    SECS_FLAG, BUILD_HKP_10  ; 1-second timer elapsed
   ; No reasons found to leave this background task, so continue ...
   sjmp  BUILD_HKP_20
BUILD_HKP_10:
   ljmp  BUILD_HKP_99                                           
BUILD_HKP_20:
$endif

   ; If more items remain in a multi-item reading, do not get more data
   ; from the housekeeping script.  Simply jump over the lines below
   ; which read from the script and continue processing the current line,
   ; which is stored in HKP_PARAMS+0, +1, +2, and +3.
   mov   A, HKP_PARAMS+3
   cjne  A, #0, BUILD_HKP_30

   ; Load current address of housekeeping script pointer into DPTR for
   ; reading the next script line ...
   mov   DPH, SCRIPT_PTR+0  ; restore MSB of pointer to housekeeping script
   mov   DPL, SCRIPT_PTR+1  ; restore LSB
   
   ; Read the next line of the script ...
   movx  A, @DPTR           ; read script opcode
   mov   HKP_PARAMS+0, A    ; store script opcode
   inc   DPTR
   movx  A, @DPTR           ; read script parameter byte 1
   mov   HKP_PARAMS+1, A    ; store param 1
   inc   DPTR
   movx  A, @DPTR           ; read script parameter byte 2
   mov   HKP_PARAMS+2, A    ; store param 2
   inc   DPTR
   movx  A, @DPTR           ; read script parameter byte 3 (the repeat count)
   mov   HKP_PARAMS+3, A    ; store param 3
   inc   DPTR               ; advance DPTR to start of next script line
   
   ; Save script pointer in SCRIPT_PTR+0:SCRIPT_PTR+1 for next iteration.
   mov   SCRIPT_PTR+0, DPH  ; store MSB of pointer to housekeeping script
   mov   SCRIPT_PTR+1, DPL  ; store LSB

BUILD_HKP_30:

   ; Start a large CASE statement based on the opcode, held in HKP_PARAMS+0
   mov   A, HKP_PARAMS+0

   ;--------------------------------------------------------------------------
BUILD_HKP_00:               ; OPCODE = 0
   cjne  A, #0, BUILD_HKP_00_END ; jump if opcode <> 0
   ; opcode=0 --> done ...
   clr   HKP_BUILD_FLAG     ; housekeeping construction is done
   setb  SEND_HKP_FLAG      ; time to send housekeeping buffer
   ljmp  BUILD_HKP_99       ; exit
BUILD_HKP_00_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_01:               ; OPCODE = 1
   cjne  A, #1, BUILD_HKP_01_END ; jump if opcode <> 1
   ; opcode=1 --> read CODE space and copy into housekeeping packet
   ; Prepare registers for call to MEM_COPY.
   setb  SRC_FLAG           ; SRC_FLAG=1 --> copy from CODE space
   sjmp  BUILD_HKP_021      ; same code as XDATA copy (except SRC_FLAG)
BUILD_HKP_01_END:
   ;--------------------------------------------------------------------------
   
   ;--------------------------------------------------------------------------
BUILD_HKP_02:               ; OPCODE = 2
   cjne  A, #2, BUILD_HKP_02_END ; jump if opcode <> 2
   ; opcode=2 --> read XDATA space and copy into housekeeping packet
   ; Prepare registers for call to MEM_COPY.
   clr   SRC_FLAG           ; SRC_FLAG=0 --> copy from XDATA space

BUILD_HKP_021:   
   ; Put Source address in R3:R2 ...
   mov   R3, HKP_PARAMS+1
   mov   R2, HKP_PARAMS+2
   ; Put Destination address in R5:R4 ...
   mov   R5, HKP_PTR+0      ; MSB
   mov   R4, HKP_PTR+1      ; LSB
   ; Number of bytes to copy in R0:R7 ...
   mov   R0, #0
   mov   R7, HKP_PARAMS+3
   ; Parameter "SRC_FLAG" was already set/cleared (above)

   ; If the source address (R3:R2) is in the I/O space, unlock/enable
   ; appropriate devices (if reading such warrants this action) ...
BUILD_HKP_0221:
   cjne  R3, #HIGH(PH_A), BUILD_HKP_0222
   setb  PHA_DIS            ; Enable reading of PH_A (by disabling ACTEL)
BUILD_HKP_0222:
   cjne  R3, #HIGH(PH_B), BUILD_HKP_0223
   setb  PHA_DIS            ; Enable reading of PH_B (by disabling ACTEL)
BUILD_HKP_0223:
   cjne  R3, #HIGH(CONTROLS2), BUILD_HKP_0224
   clr   C2_LOCK            ; Unlock CONTROLS2
BUILD_HKP_0224:

   ; Copy the values ...
BUILD_HKP_0230:
   lcall MEM_COPY
   
   ; Lock/disable PH_A, PH_B, and CONTROLS2, regardless
   ; of whether we unlocked/enabled them above (it doesn't hurt) ...
   clr   PHA_DIS            ; Disable reading PH_A & PH_B (reenable ACTEL)
   setb  C2_LOCK            ; Lock CONTROLS2

   ; update destination address -- HKP_PTR+0:HKP_PTR+1
   mov   HKP_PTR+0, R5      ; store updated HKP address MSB
   mov   HKP_PTR+1, R4      ; store updated HKP address LSB
   mov   HKP_PARAMS+3, #0   ; flag this line of the script as done.
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_02_END:
   ;--------------------------------------------------------------------------
   
   ;--------------------------------------------------------------------------
BUILD_HKP_03:               ; OPCODE = 3
   cjne  A, #3, BUILD_HKP_03_END ; jump if opcode <> 3
   ; opcode=3 --> read DATA (internal) RAM and copy into housekeeping packet
   ; Put source address in R0 for indirect addressing ...
   mov   R0, HKP_PARAMS+2
   ; Put Destination address in DPTR ...
   mov   DPH, HKP_PTR+0     ; MSB
   mov   DPL, HKP_PTR+1     ; LSB
   ; Number of bytes is in 'HKP_PARAMS+3'
BUILD_HKP_031:
   mov   A, @R0             ; read data byte (indirect read)
   movx  @DPTR, A           ; save
   inc   DPTR               ; point to next destination address
   inc   R0                 ; point to next DATA item
   djnz  HKP_PARAMS+3, BUILD_HKP_031  ; loop until done
   ; update destination address -- HKP_PTR+0:HKP_PTR+1
   mov   HKP_PTR+0, DPH     ; store updated HKP address MSB
   mov   HKP_PTR+1, DPL     ; store updated HKP address LSB
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_03_END:
   ;--------------------------------------------------------------------------

$if 0
   ; I don't believe that such an opcode is necessary.  Are there any bilevel
   ; devices for which a MUX setup is required prior to reading?
   ;--------------------------------------------------------------------------
BUILD_HKP_04:               ; OPCODE = 4
   cjne  A, #4, BUILD_HKP_04_END ; jump if opcode <> 4
   ; opcode=4 --> Digital I/O byte (w/ MUX setup)

   mov   HKP_PARAMS+3, #0   ; flag this line of the script as done.
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_04_END:
   ;--------------------------------------------------------------------------
$endif

$if 0
   ; I don't believe that such an opcode is necessary.  Are there any bilevel
   ; devices for which a MUX setup is required prior to reading?
   ;--------------------------------------------------------------------------
BUILD_HKP_05:               ; OPCODE = 5
   cjne  A, #5, BUILD_HKP_05_END ; jump if opcode <> 5
   ; opcode=5 --> Digital I/O bit (w/ MUX setup and bit mask)

   mov   HKP_PARAMS+3, #0   ; flag this line of the script as done.
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_05_END:
   ;--------------------------------------------------------------------------
$endif
   
   ;--------------------------------------------------------------------------
BUILD_HKP_06:               ; OPCODE = 6
   cjne  A, #6, BUILD_HKP_06_END ; jump if opcode <> 6
   ; opcode=6 --> Analog reading (w/ MUX setup and delay)
   jbc   HKP_RESUME, BUILD_HKP_061  ; we are RESUMING after the MUX delay

   ; Setup MUX for A/D reading ...
   mov   A, HKP_PARAMS+1    ; device MUX address for this sensor
   mov   DPTR, #MUX         ; I/O port for writing MUX address
   movx  @DPTR, A           ; write MUX address to port

BUILD_HKP_06_ENTRY_POINT:   ; entry point for other A/D opcodes

   ; If MUX delay is zero, simply continue ...
   mov   A, HKP_PARAMS+2    ; parameter 2 holds MUX settling time
   jz    BUILD_HKP_061

   ; MUX delay is not zero, so setup Timer1 to interrupt when proper
   ; delay time has expired (and set a flag for us to check) ...
   
   ; MUX delays are multiples of 0.384 msec, i.e., 256 machine cycles using
   ; an 8-MHz crystal.  To achieve this timing, Timer1's count-up register
   ; pair, TH1:TL1, is loaded with the value 10000h - mm00h, where 'mm' is
   ; the delay time found in HKP_PARAMS+2.  More simply, TL1 is zeroed and
   ; TH1 is set to the value 100h - mmh.
   ;
   ; For example, if HKP_PARAMS+2 holds the value 3 (i.e., we want a delay
   ; time for MUX settling of 3 * 0.384 msec = 1.152 msec), then TH1:TL1
   ; is loaded with 10000h - 300h = FD00h.  This value is 300h counts
   ; (768 decimal) below the rollover value of 10000h.
   ; Using our 8-MHz crystal, there are 8e6/12 machine cycles per second.
   ; Thus, a rollover count of 768 yields a delay time of 1.152 msec.

op06_latency   equ   62h    ; machine cycle delay between Timer1 ISR and
                            ; start of ADC conversion below.

   mov   TL1, #0            ; TL1 (the LSB) = 0
   clr   C                  ; prepare for SUBB operation
   clr   A
   subb  A, HKP_PARAMS+2
   mov   TH1, A             ; MSB holds (100h - delay parameter)
   mov   TL1, #op06_latency ; account for latency
   clr   TIMER1_DONE        ; flag used for housekeeping background task
   clr   TIMER1_EDGE        ; allow the CRC background task some time
   setb  TR1                ; start Timer1 (should already be)
   setb  ET1                ; enable Timer1 interrupt (should already be)

   ; Set the logical bit "HKP_RESUME" to tell us that next pass we are
   ; merely resuming this script command where we left off ...
   setb  HKP_RESUME         ; 1=we are beginning a MUX delay (i.e., we
                            ; will interrupt this script and must RESUME it
                            ; later)
   ljmp  BUILD_HKP_99       ; exit houskeeping building until Timer1 indicates
                            ; that MUX delay is done (TIMER1_EDGE='1').

BUILD_HKP_061:

   ; Delay 20 us for MUX settling just in case an A/D delay time of zero was
   ; specified (we need some delay time) ...
   mov   R0, #7             ; loop counter
   djnz  R0, $

   ; With MUX delay done, start the ADC by writing ANYTHING to
   ; address 'ADC_SPACE' ...
   mov   DPTR, #ADC_SPACE   ; write anything to start A/D conversion
   movx  @DPTR, A
   
   ; Delay 20 us for A/D to complete its conversion ...
   mov   R0, #7             ; loop counter
   djnz  R0, $
   
   ; Read the A/D (from same address as used to start the conversion) ...
   movx  A, @DPTR           ; read the analog value
   
   ; Write the sensor value to housekeeping packet ...
   mov   DPH, HKP_PTR+0     ; recall HKP address
   mov   DPL, HKP_PTR+1
   movx  @DPTR, A           ; write the value to HKP
   inc   DPTR               ; advance HKP DPTR
   mov   HKP_PTR+0, DPH     ; store updated HKP address MSB
   mov   HKP_PTR+1, DPL     ; store updated HKP address LSB

   ; Prepare for next A/D reading by advancing HKP_PARAMS values ...
   inc   HKP_PARAMS+1       ; increment MUX address
   dec   HKP_PARAMS+3       ; decrement Number of Bytes to read
   ; Opcode and delay time parameters remain unchanged.

   clr   TIMER1_EDGE        ; allow the CRC background task some time

   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_06_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_07:               ; OPCODE = 7
   cjne  A, #7, BUILD_HKP_07_END ; jump if opcode <> 7
   ; opcode=7 --> move housekeeping pointer to any absolute location within
   ; the housekeeping buffer.

BUILD_HKP_07_ENTRY_POINT:

   ; Ensure the requested housekeeping pointer is in the range TS17 to
   ; TS17+03FFh, inclusive.  To do this, subtract TS17 from the desired
   ; address and ensure the result is in the range 000-3FFh (i.e., that
   ; resulting MSB is less than 4) ...
   clr   C                  ; prepare for SUBB operation
   mov   A, HKP_PARAMS+2    ; LSB of requested address
   subb  A, #LOW(TS17)      ; no need to store value
   mov   A, HKP_PARAMS+1    ; MSB of requested address
   subb  A, #HIGH(TS17)
   jc    BUILD_HKP_071      ; Carry=1 --> new address LESS than TS17 (too low)
   subb  A, #4              ; verify MSB is less than 04h
   jnc   BUILD_HKP_071      ; Carry=0 --> new address too high
   ; new address is OK, so store it ...
   mov   HKP_PTR+0, HKP_PARAMS+1  ; store new MSB
   mov   HKP_PTR+1, HKP_PARAMS+2  ; store new LSB
BUILD_HKP_071:
   mov   HKP_PARAMS+3, #0   ; flag this line of the script as done.
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_07_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_08:               ; OPCODE = 8
   cjne  A, #8, BUILD_HKP_08_END ; jump if opcode <> 8
   ; opcode=8 --> skip space in the housekeeping buffer.  This is a relative
   ; change in the housekeeping buffer pointer, advancing it by the number of
   ; bytes found in HKP_PARAMS+3.
   
   ; In order to take advantage of the range checking performed by opcode 7
   ; (absolute pointer move), we will use that same code by changing the
   ; supplied parameters from relative to absolute and then call opcode 7.
   mov   A, HKP_PARAMS+3    ; number of byte to move ahead in buffer
   add   A, HKP_PTR+1       ; add LSB of current buffer pointer
   mov   HKP_PARAMS+2, A    ; store LSB as HKP_PARAMS+2 (absolute address LSB)
   mov   A, HKP_PTR+0
   addc  A, #0              ; propagate carry bit to MSB
   mov   HKP_PARAMS+1, A    ; store MSB as HKP_PARAMS+1 (absolute address MSB)
   ljmp  BUILD_HKP_07_ENTRY_POINT ; jump to opcode 7 processing
BUILD_HKP_08_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_09:               ; OPCODE = 9
   cjne  A, #9, BUILD_HKP_09_END ; jump if opcode <> 9
   ; opcode=9 --> zero-out space in the housekeeping buffer
   mov   DPH, HKP_PTR+0     ; pointer into housekeeing buffer
   mov   DPL, HKP_PTR+1
   clr   A
BUILD_HKP_091:
   movx  @DPTR, A           ; clear current location
   inc   DPTR
   djnz  HKP_PARAMS+3, BUILD_HKP_091  ; loop until done
   ; update destination address -- HKP_PTR+0:HKP_PTR+1
   mov   HKP_PTR+0, DPH     ; store updated HKP address MSB
   mov   HKP_PTR+1, DPL     ; store updated HKP address LSB
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_09_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_0A:               ; OPCODE = 0Ah
   cjne  A, #0Ah, BUILD_HKP_0A_END ; jump if opcode <> 0Ah
   ; opcode=0Ah --> NOP (merely a placeholder in the script for future use)
   mov   HKP_PARAMS+3, #0   ; flag this line of the script as done.
   ljmp  BUILD_HKP_NEXT     ; loop to top
BUILD_HKP_0A_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_46:               ; OPCODE = 46 (hex)
   cjne  A, #46h, BUILD_HKP_46_END ; jump if opcode <> 46h
   ; opcode=46 (hex) --> TDC 'A' X-Analog reading (w/ 2 MUX setups)
   jbc   HKP_RESUME, BUILD_HKP_461  ; we are RESUMING after the MUX delay

   ; Setup TDC MUX by writing out 3-wire interface ...
   ; Put TDC MUX select value in R0.
   mov   R0, HKP_PARAMS+1
   lcall TDC_A              ; write R0:R1 to TDC 'A' (R1 irrelevant)

   ; Setup MUX for A/D reading ...
   mov   A, HKP_PARAMS+0    ; device MUX address for this sensor
   mov   DPTR, #MUX         ; I/O port for writing MUX address
   movx  @DPTR, A           ; write MUX address to port

   ; opcode 6 has all the A/D setup and delay code, so simply jump to that ...
   ljmp  BUILD_HKP_06_ENTRY_POINT

BUILD_HKP_461:
   ; opcode 6 has all the A/D setup and delay code, so simply jump to that ...
   ljmp  BUILD_HKP_061
   
BUILD_HKP_46_END:
   ;--------------------------------------------------------------------------

   ;--------------------------------------------------------------------------
BUILD_HKP_82:               ; OPCODE = 82 (hex)
   cjne  A, #82h, BUILD_HKP_82_END ; jump if opcode <> 82h
   ; opcode=82 (hex) --> TDC 'B' X-Analog reading (w/ 2 MUX setups)
   jbc   HKP_RESUME, BUILD_HKP_821  ; we are RESUMING after the MUX delay

   ; Setup TDC MUX by writing out 3-wire interface ...
   ; Put TDC MUX select value in R0.
   mov   R0, HKP_PARAMS+1
   lcall TDC_B              ; write R0:R1 to TDC 'B' (R1 irrelevant)

   ; Setup MUX for A/D reading ...
   mov   A, HKP_PARAMS+0    ; device MUX address for this sensor
   mov   DPTR, #MUX         ; I/O port for writing MUX address
   movx  @DPTR, A           ; write MUX address to port

   ; opcode 6 has all the A/D setup and delay code, so simply jump to that ...
   ljmp  BUILD_HKP_06_ENTRY_POINT

BUILD_HKP_821:
   ; opcode 6 has all the A/D setup and delay code, so simply jump to that ...
   ljmp  BUILD_HKP_061
   
BUILD_HKP_82_END:
   ;--------------------------------------------------------------------------

   ; If we have an improper script opcode, we will reach this point
   ; and treat it as an end-of-script condition (opcode=0) ...
   clr   HKP_BUILD_FLAG     ; housekeeping construction is done
   setb  SEND_HKP_FLAG      ; time to send housekeeping buffer
   ljmp  BUILD_HKP_99       ; exit
;   mov   HKP_PARAMS+3, #0   ; flag this line of the script as done.
;   ljmp  BUILD_HKP_NEXT     ; loop to top

$if (STAY_IN_BUILD_HKP=0)
BUILD_HKP_NEXT:
$endif

BUILD_HKP_99:
   ret                      ; BUILD_HKP
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- SEND_DUMP_PKT, SEND_HKP_PKT
; Purpose -- Entry points to routines that send 512 32-bit words of memory
;            dump data or housekeeping data.  Only 512 16-bit words of data
;            are pointed to.  The other 16 bits of each long word consist of
;            the address of the word being sent.
; Inputs -- none
; Outputs -- none
; Regisers used -- R0, R1, DPTR
; Timing -- 22,542 machine cycles (33.813 msec at 8 MHz)
; Testing -- Tested by simulation and working, 9/21/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
SEND_DUMP_PKT:
   mov   DPTR, #TS01        ; point to the dump packet block of data
   sjmp  SEND_DUMP_PKT_10   ; jump to sending loop

SEND_HKP_PKT:
   mov   DPTR, #TS17        ; point to the housekeeping packet block of data

   ; Send 512 words (1024 bytes).  Loop 512 times ...
SEND_DUMP_PKT_10:
   mov   R0, #HIGH(512)     ; loop counter MSB
   mov   R1, #LOW(512)      ; loop counter LSB
SEND_DUMP_PKT_30:
   lcall SEND_PKT           ; DPTR points to value
   inc   DPTR               ; SEND_PKT sent two bytes of data so we advance
   inc   DPTR               ;    pointer by two.
   djnz  R1, SEND_DUMP_PKT_30   ; decrement counter LSB
   djnz  R0, SEND_DUMP_PKT_30   ; decrement counter MSB

SEND_DUMP_PKT_99:
   ret                      ; SEND_DUMP_PKT & SEND_HKP_PKT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- SEND_PKT
; Purpose -- Send one 4-byte packet from COS DCE to CS computer
; Called by -- SEND_DUMP_PKT & SEND_HKP_PKT
; Inputs -- DPTR points to 16-bit value to send
; Outputs -- none
; Timing -- 38 machine cycles (0.057 msec @ 8-MHz)
; Registers used -- A, DPTR
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
SEND_PKT:
   mov   HKPBUFF+2, DPL     ; Address LSB stored while changing DPTR
   mov   HKPBUFF+3, DPH     ; Address MSB stored while changing DPTR
   movx  A, @DPTR           ; Fetch the data LSB
   mov   HKPBUFF+0, A       ; save LSB
   inc   DPTR               ; point to upper data byte
   movx  A, @DPTR           ; Fetch the data MSB
   mov   HKPBUFF+1, A       ; save MSB

   ; The transmitter is four byte registers.  When a value is written to
   ; address 5300h, the trasmitter with start shifting out the 32 bit word.
   ; Bytes may be written to the registers in any order, so long as the one
   ; at address 5300h is written last.  The housekeeping address is written
   ; to 5300:5200, the data written to 5100:5000.
   
   mov   A, HKPBUFF+0       ; fetch data LSB
   mov   DPTR, #TXD_HARDWARE  ; base transmit register (5000h)
   movx  @DPTR, A           ; write data LSB
   mov   A, HKPBUFF+1       ; fetch data MSB
   inc   DPH                ; 5100h is an xmit register
   movx  @DPTR, A           ; write data MSB
   mov   A, HKPBUFF+2       ; fetch address LSB
   inc   DPH                ; 5200h is an xmit register
   movx  @DPTR, A           ; write address LSB
   mov   A, HKPBUFF+3       ; fetch address MSB
   anl   A, #1Fh            ; map to 0000h..1FFFh
   inc   DPH                ; 5300h is MSB xmit register
   movx  @DPTR, A           ; write address MSB and cause transfer to begin

   ; Restore DPTR which points to word just transmitted
   mov   DPL, HKPBUFF+2     ; restore pointer
   mov   DPH, HKPBUFF+3     ; restore pointer

SEND_PKT_99:
   ret                      ; SEND_PKT
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- SAFE_CONTROLS
; Purpose -- Places the hardware in a known, safe state.
; Inputs -- none
; Outputs -- none
; Timing -- 30 machine cycles (0.045 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/21/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
SAFE_CONTROLS:

   ; 'RAM_BANK' is the most-significant address bit of DCE-B RAM chip U8.
   ; Set it to '1' and don't change throughout the code ...
   setb  RAM_BANK

   ; Though Boot does not read PH data, we will enable the ACTEL devices which
   ; collect it ...
   clr   PHA_DIS            ; enable ACTEL to gather and store PH data

   ; Reset the ACTEL devices on the TDCs and DCE-A by giving them a low pulse
   ; of at least 10us ...
   clr   ACTEL_RESET        ; active low reset
   mov   R0, #3             ; Loop for 10 us
   djnz  R0, $
   setb  ACTEL_RESET        ; enable the ACTEL devices on the TDCs and DCE-A

   ; Turn off (reset to zero) all bits of CONTROLS1 and CONTROLS2 ...
   clr   C2_LOCK            ; make sure CONTROLS2 is unlocked
   mov   A, #0
   mov   DPTR, #CONTROLS1
   movx  @DPTR, A           ; write 00h to CONTROLS1 register
   mov   DPTR, #CONTROLS2
   movx  @DPTR, A           ; write 00h to CONTROLS2 register
   setb  C2_LOCK            ; make sure CONTROLS2 is locked
   
   ; Initialize the 3-wire interface to a safe state ...
   setb  TDC_CLOCK          ; P1.0, TDC 'A' & 'B' clock
   setb  TDC_DATA           ; P1.1, TDC 'A' & 'B' data out
   clr   TDC_A_ENABLE       ; P1.2, TDC 'A' Active high enable
   clr   TDC_B_ENABLE       ; P1.3, TDC 'B' Active high enable
   
SAFE_CONTROLS_99:
   ret                      ; SAFE_CONTROLS
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- INIT_RAM
; Purpose -- Clears internal & external RAM.
; Inputs -- none
; Outputs -- Internal RAM:  Registers 20h up to the base of the internal
;               STACK are cleared.
;            External RAM:  Upload packet (1024 bytes), Download packet (1024)
;               Housekeeping packet (1024), and Command packet (28 bytes)
; Timing -- Internal RAM:     215 machine cycles (0.323 msec @ 8-MHz)
;           External RAM:  20,230 machine cycles (30.35 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
INIT_RAM:
   ; Clear internal RAM (DSEG and BSEG) up to bottom of the STACK ...
   clr   C
   mov   A, #_STACK+1       ; first address used by STACK segment
   subb  A, #20h            ; first address used by general purpose RAM
   mov   R7, A              ; number of bytes to clear
   mov   R0, #20h
INIT_RAM_00:
   mov   @R0, #0
   inc   R0
   djnz  R7, INIT_RAM_00
   
   ; Clear the upload data area (1024 bytes starting at address RS01)
   ; using the subroutine 'CLEAR_RAM' ...
   mov   DPTR, #RS01
   mov   R6, #HIGH(1024)
   mov   R7, #LOW(1024)
   lcall CLEAR_RAM

   ; Clear the command packet area (28 bytes starting at address RS17)
   ; using the subroutine 'CLEAR_RAM' ...
   mov   DPTR, #RS17
   mov   R6, #HIGH(28)
   mov   R7, #LOW(28)
   lcall CLEAR_RAM

   ; Clear the download data area (1024 bytes starting at address TS01)
   ; using the subroutine 'CLEAR_RAM' ...
   mov   DPTR, #TS01
   mov   R6, #HIGH(1024)
   mov   R7, #LOW(1024)
   lcall CLEAR_RAM

   ; Clear the housekeeping packet area (1024 bytes starting at address TS17)
   ; using the subroutine 'CLEAR_RAM' ...
   mov   DPTR, #TS17
   mov   R6, #HIGH(1024)
   mov   R7, #LOW(1024)
   lcall CLEAR_RAM

   ; Clear the scratch variable space in XDATA (256 bytes starting at PC00)
   ; using the subroutine 'CLEAR_RAM' ...
   mov   DPTR, #PC00
   mov   R6, #HIGH(256)
   mov   R7, #LOW(256)
   lcall CLEAR_RAM

INIT_RAM_99:
   ret                      ; INIT_RAM   
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CLEAR_RAM
; Purpose -- Clears a specified range of External RAM (XDATA)
; Inputs -- source address in DPTR (first address cleared)
;           number of bytes = R6:R7
; Outputs -- Specified XDATA memory set to zero.
; Timing -- For number of bytes = N
;           Timing = 6N machine cycles (plus small looping overhead)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CLEAR_RAM:
   ; fixup number of bytes if LSB is non-zero (to make DJNZ commands work)
   cjne  R7, #0, CLEAR_RAM_10
   sjmp  CLEAR_RAM_30
CLEAR_RAM_10:
   inc   R6                 ; if LSB<>0 we must increment MSB for looping

CLEAR_RAM_30:
   clr   A                  ; value to write to memory is zero

CLEAR_RAM_40:
   movx  @DPTR, A           ; write byte to XDATA
   inc   DPTR               ; advance destination pointer and re-save it
   djnz  R7, CLEAR_RAM_40   ; loop by decrementing counter LSB
   djnz  R6, CLEAR_RAM_40   ; loop by decrementing counter MSB

CLEAR_RAM_99:
   ret                      ; CLEAR_RAM
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- COPY_PATCHABLE
; Purpose -- Copies patchable constants from CODE space to XDATA.
; Inputs -- table 'patch_table' specifies the source address (16-bits), the
;           destination address (16-bits), and the number of bytes (16-bits).
; Outputs -- none
; Timing -- Defining:  BLOCKS = number of patchable constant blocks to copy
;                      BYTES = total number of bytes copied
;                      TIMING = total time, measured in machine cycles
;           Then:      TIMING = (29 * BYTES) + (48 * BLOCKS) + 38
;           Default case:  TIMING = 37,816 machine cycles (56.7 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
COPY_PATCHABLE:

   mov   DPTR, #patch_table
COPY_PATCHABLE_00:
   ; Read SOURCE address and store in R3:R2 for MEM_COPY, below ...
   clr   A                  ; prepare for MOVC instruction
   movc  A, @A+DPTR         ; read address MSB
   mov   R3, A              ; store source address = R3:R2
   inc   DPTR
   clr   A
   movc  A, @A+DPTR         ; read address LSB
   mov   R2, A              ; source LSB
   inc   DPTR
   ; Read DESTINATION address and store in R5:R4 for MEM_COPY, below ...
   clr   A                  ; prepare for MOVC instruction
   movc  A, @A+DPTR         ; read address MSB
   mov   R5, A              ; store destination address = R5:R4
   inc   DPTR
   clr   A
   movc  A, @A+DPTR         ; read address LSB
   mov   R4, A              ; destination LSB
   inc   DPTR
   ; Read Number of Bytes and store in R0:R7 for MEM_COPY, below ...
   clr   A
   movc  A, @A+DPTR
   mov   R0, A              ; store number of bytes = R0:R7
   inc   DPTR
   clr   A
   movc  A, @A+DPTR
   mov   R7, A              ; store LSB of number of bytes
   inc   DPTR               ; prepare for next iteration
   cjne  R0, #0, COPY_PATCHABLE_10  ; OK if nonzero
   cjne  R7, #0, COPY_PATCHABLE_10  ; OK if nonzero
   sjmp  COPY_PATCHABLE_99  ; done
   
COPY_PATCHABLE_10:
   push  DPH                ; save DPTR, which points to 'patch_table'
   push  DPL
   setb  SRC_FLAG           ; source memory space = SRC_FLAG (0=XDATA, 1=CODE)
   lcall MEM_COPY
   pop   DPL                ; restore DPTR for next copy area
   pop   DPH
   sjmp  COPY_PATCHABLE_00  ; look for next block to copy

COPY_PATCHABLE_99:
   ret                      ; COPY_PATCHABLE
;-----------------------------------------------------------------------------
   
;*****************************************************************************
; Subroutine -- REPORT_DIAGNOSTIC
; Purpose -- Generates diagnostic and possibly HST error.
; Inputs -- Diagnostic code in ACC
; Outputs -- Diagnostic on 32-level push-down stack.  HST error and parameter
;            (if applicable) on 8-level error stack and parameter stack.
; Registers -- R0, R1, ACC, B, PSW, DPTR
; Timing --    871 machine cycles for diagnostic alone (1.31 msec @ 8-MHz)
;           ~1,027 machine cycles for diag. + HST error (1.54 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
REPORT_DIAGNOSTIC:

   ; Only report diagnostics as HST errors if the table 'HSTERR_TABLE' has
   ; a non-null pointer at the 2-byte location corresponding to the
   ; diagnostic code specified (i.e., at the WORD beginning at address
   ; 'HSTERR_TABLE + 2*ACC')
   push  ACC                   ; save diagnostic code for later use
   mov   DPTR, #HSTERR_TABLE   ; table of HST error codes
   lcall XDATA_TLU_WORD        ; 2-byte look-up value returned in B:A
   ; if either A<>0 or B<>0, this is a non-null pointer and we save this
   ; diagnostic as an HST error ...
   jnz   REPORT_DIAGNOSTIC_10  ; non-null --> save the HST error
   mov   A, B
   jnz   REPORT_DIAGNOSTIC_10  ; non-null --> save the HST error
   sjmp  REPORT_DIAGNOSTIC_50  ; null pointer --> save only as a diagnostic

REPORT_DIAGNOSTIC_10:
   ; Note:  The diagnostic code resides on the 8051 stack.

   ; Save the diagnostic as an HST error.

   ; Look for the first unfilled location of the HST error buffer.  An
   ; unfilled location is a zero value.  If the buffer is filled, the
   ; last (8th) HST error is overwritten.
   mov   R0, #8             ; max number of hst errors allowed per second
   mov   R1, #-1            ; count-up value
   mov   DPTR, #mLFDERR-1   ; point 1 byte prior to HST error array

REPORT_DIAGNOSTIC_20:
   inc   DPTR               ; increment pointe into HST error array
   inc   R1
   movx  A, @DPTR           ; retrieve value held in HST error array
   jz    REPORT_DIAGNOSTIC_30 ; space available when zero
   djnz  R0, REPORT_DIAGNOSTIC_20 ; keep looking until we hit the end of list
   
   ; If more than 8 HST errors occurred during this 1-second period,
   ; we throw away the additional ones ...
   sjmp  REPORT_DIAGNOSTIC_50  ; save only as a diagnostic
   
REPORT_DIAGNOSTIC_30:
   pop   ACC                ; retrieve diagnostic code from 8051 stack
   movx  @DPTR, A           ; write diagnostic code to HST error array

REPORT_DIAGNOSTIC_40:
   ; Note:  The diagnostic code DOES NOT reside on the 8051 stack.  Only
   ;        in ACC.

   ; Perform a table look-up to find the HST error code associated
   ; with this diagnostic ...
   push  ACC                ; save diagnostic code for later use

   mov   DPTR, #HSTERR_TABLE ; table of HST error codes
   lcall XDATA_TLU_WORD      ; 2-byte look-up value returned in B:A
   mov   DPH, B
   mov   DPL, A
   clr   A                  ; zero index
   lcall XDATA_TLU_WORD     ; 2-byte look-up value returned in B:A
   push  ACC                ; push parameter MSB onto 8051 stack
   push  B                  ; push parameter LSB onto 8051 stack
   mov   DPTR, #mLFDERRP    ; base address of error parameter buf. 'mLFDERRP'
   mov   A, R1              ; offset into 'mLFDERRP' (from above)
   rl    A                  ; double it for word offsets (vice byte offsets)
   add   A, DPL             ; add offset to DPTR
   mov   DPL, A
   mov   A, DPH
   addc  A, #0              ; propagate the carry bit
   mov   DPH, A
   pop   ACC                ; pop parameter LSB off 8051 stack
   movx  @DPTR, A           ; Store LSB in location 'mLFDERRP'+2*R1
   inc   DPTR
   pop   ACC                ; pop parameter MSB off 8051 stack
   movx  @DPTR, A           ; Store MSB in location 'mLFDERRP'+2*R1 + 1

REPORT_DIAGNOSTIC_50:
   ; Note:  The diagnostic code resides on the 8051 stack.
   
   ; Put this new diagnostic on the 32-element push-down stack.
   
   ; Using (i) as subscript notation, we will put the most-recent
   ; diagnostic -- "diagnostic(i)" -- and the serial number corresponding
   ; to it -- "serial_number(i)" -- on the 8051 stack.  Older elements on
   ; the push-down stack are subscripted (i-1), (i-2), etc.

   ; "diagnostic(i)" still resides on the 8051 stack.
   pop   ACC                ; retrieve "diagnostic(i)" from 8051 stack
   mov   B, A               ; save "diagnostic(i)" temporarily

   ; First, update parameter 'mDIAG_NUM', the highest diagnostic serial number
   mov   DPTR, #mDIAG_NUM
   movx  A, @DPTR           ; retrieve "serial_number(i-1)"
   inc   A                  ; increment serial number --> "serial_number(i)"
   movx  @DPTR, A           ; store it
   
   ; Prepare for looping operation below ...
   push  ACC                ; push "serial_number(i)" on 8051 stack
   push  B                  ; push "diagnostic(i)" on 8051 stack

   ; Begin the push-down process and repeat 32 times ...
   mov   DPTR, #mLFDDIAGS   ; point to top of push-down stack
   mov   R0, #32            ; number of push-down operations
REPORT_DIAGNOSTIC_60:   
   movx  A, @DPTR           ; read "diagnostic(i-1)" from push-down
   mov   B, A               ; save it temporarily
   pop   ACC                ; retrieve "diagnostic(i)" from 8051 stack
   movx  @DPTR, A           ; put "diagnostic(i)" on push-down stack
   inc   DPTR               ; point to "serial_number(i-1)" on push-doan
   movx  A, @DPTR           ; read "serial_number(i-1)" from push-down
   mov   R1, A              ; save it temporarily
   pop   ACC                ; retrieve "serial_number(i)" from 8051 stack
   movx  @DPTR, A           ; put "serial_number(i)" on push-down stack
   mov   A, R1              ; mov "serial_number(i-1)" to ACC
   push  ACC                ; push "serial_number(i-1)" on 8051 stack
   push  B                  ; push "diagnostic(i-1)" on 8051 stack
   inc   DPTR               ; prepare for next loop
   djnz  R0, REPORT_DIAGNOSTIC_60  ; decrement loop counter and repeat
   
   ; Pop the last elements off the 8051 stack and throw them away.
   pop   ACC
   pop   ACC

   ; Increment the number of diagnostics stored this 1-second interval ...
   mov   DPTR, #mCURR_NUM
   movx  A, @DPTR
   inc   A
   movx  @DPTR, A

REPORT_DIAGNOSTIC_99:
   ret                      ; REPORT_DIAGNOSTIC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- DOCRC
; Purpose -- Compute the CRC for the next byte.
; Inputs -- Current CRC in CRCH:CRCL; Byte to process in ACC.
; Outputs -- Updated value to CRC in CRCH:CRCL.
; Timing -- 33 machine cycles per byte (0.050 msec @ 8-MHz)
; Testing -- 
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
DOCRC:
   mov   TEMP, A            ; save the received byte
   mov   A, CRCL            ; swap high and low
   mov   DXH, A
   mov   A, CRCH
   xrl   A, TEMP            ; XOR with the data
   mov   DXL, A
   swap  A
   anl   A, #0FH            ; shift right 4
   xrl   A, DXL
   mov   DXL, A
   swap  A
   anl   A, #0F0H           ; shift left 4
   xrl   A, DXH
   mov   DXH, A
   mov   A, DXL
   swap  A
   mov   TEMP1, A
   anl   A, #0F0H
   clr   C
   rlc   A
   xch   A, TEMP1
   anl   A, #0FH
   rlc   A                  ; shift left 5
   xrl   A, DXH
   mov   CRCH, A
   mov   A, TEMP1
   xrl   A, DXL
   mov   CRCL, A
   mov   A, TEMP            ; restore the data byte
   ret                      ; DOCRC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- BG_CRC (Background CRC computations)
; Purpose -- Computes CRC over a range of memory (XDATA or CODE).
; Input -- 'MEMPTR' points to next memory address to process (big-endian).
;          'CRC_LEN' is number of bytes remaining to compute (0 = 65536)
;          'REGION_FLAG' is memory specifier (0=XDATA, 1=CODE)
; Output -- Computed CRC in 'mLFMCRC'
; Timing -- Test case:  N=1024 bytes
;              w/o interruptions:  52,336 machine cycles (78.5 msec @ 8-MHz)
;                  Worst Case (N=65,536) --> 5.02 seconds @ 8-MHz
;              w/  interruptions:  Add 138 machine cycles per (timer)
;                  interrupt (0.207 msec per interruption @ 8-MHz).
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
$SET (STAY_IN_BG_CRC=1)

FILL_TO_BOUNDARY   16, 0FFh
BG_CRC:
   push  PSW                ; save register bank in use
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   ; Retrieve the background copy of CRC ...
   mov   CRCH, CRCHa
   mov   CRCL, CRCLa
   
   mov   DPH, MEMPTR+0      ; address of next byte to read (big-endian)
   mov   DPL, MEMPTR+1
BG_CRC_10:

$if STAY_IN_BG_CRC
   ; Reasons to leave this background task prior to completion ...
   jb    RXD_FLAG,    BG_CRC_98  ; opcode received
   jb    SECS_FLAG,   BG_CRC_98  ; 1-second timer tick occured
   jb    TIMER1_EDGE, BG_CRC_98  ; Housekeeping A/D MUX delay expired
   ; No reasons found to leave this background task, so continue ...
$endif
   
   ; read from XDATA or CODE?
   jnb   REGION_FLAG, BG_CRC_20  ; jump if XDATA space (REGION_FLAG=0)
   clr   A
   movc  A, @A+DPTR
   sjmp  BG_CRC_30
BG_CRC_20:
   movx  A, @DPTR

BG_CRC_30:
   lcall DOCRC              ; update CRC with this data byte
   inc   DPTR               ; advance pointer to next byte
   mov   MEMPTR+0, DPH      ; store address, in case we must leave BG task
   mov   MEMPTR+1, DPL      ; ditto

$if STAY_IN_BG_CRC
   djnz  CRC_LEN+1, BG_CRC_10 ; decrement counter LSB
   djnz  CRC_LEN+0, BG_CRC_10 ; decrement counter MSB
$else
   djnz  CRC_LEN+1, BG_CRC_98 ; decrement counter LSB
   djnz  CRC_LEN+0, BG_CRC_98 ; decrement counter MSB
$endif

   ; we're done, so clean up and exit ...
   mov   CRCHa, CRCH
   mov   CRCLa, CRCL
   lcall STORE_CRC          ; store the answer (little-endian)
                            ; return value:  ACC = 'next region'
   mov   REGION_CRC, A      ; store the 'next region'
   lcall INIT_CRC           ; initialize variables for the new region
   sjmp  BG_CRC_99          ; exit

BG_CRC_98:
   ; Store CRCH:CRCL in the registers CRCHa:CRCLa so that foreground calls to
   ; "DOCRC" will not destroy them.
   mov   CRCHa, CRCH
   mov   CRCLa, CRCL

BG_CRC_99:
   pop   PSW                ; restore register bank to previous
   ret                      ; BG_CRC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- INIT_CRC
; Purpose -- Initialize CRC calculations for a new region of memory.
; Inputs -- next region number in byte-parameter 'REGION_CRC'
; Outputs -- variables 'MEMPTR', 'CRC_LEN', 'CRCHa', and 'CRCLa' initialized
; Timing -- 72 machine cycles (0.108 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/13/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
INIT_CRC:
   clr   CRC_FLAG           ; disable background CRC calculations

   clr   C                  ; prepare for SUBB operation
   mov   A, REGION_CRC      ; ensure the region requested is valid
   subb  A, #num_regions    ; maximum number of regions allocated
   jnc   INIT_CRC_99        ; invalid region specifier -- exit

INIT_CRC_10:
   mov   DPTR, #CRC_REGIONS ; base address of CRC regions
   mov   A, REGION_CRC      ; region number
   mov   B, #sz_regions     ; size of each region specifier
   mul   AB                 ; offset into 'CRC_REGIONS' table in B:A
   add   A, DPL             ; add offset to DPTR
   mov   DPL, A
   mov   A, B
   addc  A, DPH
   mov   DPH, A
   ; Read data from 'CRC_REGIONS' table and put in local variables ...
   movx  A, @DPTR
   mov   MEMPTR+0, A        ; MSB of start address
   inc   DPTR
   movx  A, @DPTR
   mov   MEMPTR+1, A        ; LSB of start address
   inc   DPTR
   movx  A, @DPTR
   mov   CRC_LEN+0, A       ; MSB of length
   inc   DPTR
   movx  A, @DPTR
   mov   CRC_LEN+1, A       ; LSB of length
   inc   DPTR
   movx  A, @DPTR           ; memory specifier (0=XDATA, 1=CODE)
   clr   REGION_FLAG        ; assume XDATA region
   jz    INIT_CRC_20        ; parameter OK
   setb  REGION_FLAG        ; assume CODE region
   dec   A
   jnz   INIT_CRC_99        ; parameter error -- exit

INIT_CRC_20:
   ; initialize CRC for the background task ...
   mov   CRCHa, #0FFh
   mov   CRCLa, #0FFh
   
   ; Fixup "number of bytes" if LSB is non-zero (to make DJNZ commands work)
   mov   A, CRC_LEN+1       ; length LSB
   jz    INIT_CRC_30        ; MSB is OK since LSB=0
   inc   CRC_LEN+0          ; if LSB<>0 we must increment MSB for looping
INIT_CRC_30:

   setb  CRC_FLAG           ; enable background CRC calculations

INIT_CRC_99:
   ret                      ; INIT_CRC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- STORE_CRC
; Purpose -- Saves the calculated CRC in the designated location for the
;            current CRC region.
; Inputs -- CRC in CRCHa:CRCLa, current region specifier in 'REGION_CRC'
; Outputs -- CRCHa:CRCLa written to designated location in CRC table.
;            Return value is the 'next region' in ACC.
; Timing -- 77 machine cycles (0.116 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/13/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
STORE_CRC:
   mov   DPTR, #CRC_REGIONS+5 ; 'next region' specifier of first region
   mov   A, REGION_CRC        ; region number
   mov   B, #sz_regions       ; size of each region specifier
   mul   AB                   ; offset into 'CRC_REGIONS' table in B:A
   add   A, DPL               ; add offset to DPTR
   mov   DPL, A
   mov   A, B
   addc  A, DPH
   mov   DPH, A
   ; Read 'next region' from table ...
   movx  A, @DPTR
   push  ACC                ; temporary storage
   inc   DPTR               ; point to LSB of CRC
   mov   A, CRCLa
   movx  @DPTR, A           ; store LSB of CRC
   inc   DPTR               ; point to MSB of CRC
   mov   A, CRCHa
   movx  @DPTR, A           ; store MSB of CRC
   inc   DPTR
   ; Compare new CRC with previous value (if non-zero) and send a diagnostic
   ; if requested ...
   ; Read the previous CRC, push on stack, and store new CRC as 'previous'
   ; (in preparation for the next time we work on this region) ...
   mov   B, #0              ; use B as a temporary flag register
   movx  A, @DPTR           ; retrieve previous CRC LSB
   push  ACC
   mov   A, CRCLa
   movx  @DPTR, A           ; store new CRC LSB
   inc   DPTR
   pop   ACC                ; need to check previous CRC for zero
   push  ACC                ; save it again
   jz    STORE_CRC_10
   mov   B, #1              ; B non-zero means we must compare CRCs
STORE_CRC_10:
   movx  A, @DPTR           ; retrieve previous CRC MSB
   push  ACC
   mov   A, CRCHa
   movx  @DPTR, A           ; store new CRC MSB
   inc   DPTR
   pop   ACC                ; need to check previous CRC for zero
   push  ACC                ; save it again
   jz    STORE_CRC_20
   mov   B, #1              ; B non-zero means we must compare CRCs
STORE_CRC_20:
   mov   A, B               ; move our flag register (B) into ACC to check
   jnz   STORE_CRC_30       ; if B<>0, we must compare CRCs
   pop   ACC                ; remove the 'previous CRC' from stack
   pop   ACC
   sjmp  STORE_CRC_98       ; exit

   ; we must compare the CRCs
STORE_CRC_30:
   pop   ACC                ; retrieve previous CRC MSB
   cjne  A, CRCHa, STORE_CRC_40  ; jump on miscompare
   pop   ACC                ; retrieve previous CRC LSB
   cjne  A, CRCLa, STORE_CRC_50  ; jump on miscompare
   sjmp  STORE_CRC_98       ; CRCs are equal -- exit

   ; miscompare on first byte -- pop other byte off stack
STORE_CRC_40:
   pop   ACC                ; remove second part of CRC from stack
   
STORE_CRC_50:
   ; CRCs are different.  If the MSB of the diagnostic parameter in this
   ; CRC table entry is 5Ah, then we send the LSB as the diagnostic messge.
   movx  A, @DPTR           ; retrieve MSB of diagnostic message
   cjne  A, #5Ah, STORE_CRC_98  ; no diagnostic to send -- exit
   ; must send the diagnostic ...
   inc   DPTR
   movx  A, @DPTR           ; diagnostic number is LSB
   lcall REPORT_DIAGNOSTIC  ; Record this diagnostic event

STORE_CRC_98:
   pop   ACC                ; recover the 'next region'

STORE_CRC_99:
   ret                      ; STORE_CRC
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- CLEAR_OLD_CRCS
; Purpose -- Zeros out all 'previous CRC' entries in the CRC_REGIONS table,
;            thereby preventing CRC comparison errors on the first pass.
; Inputs -- none
; Outputs -- All 'previous CRC' entries in the CRC_REGIONS table are set to
;            zero.
; Timing -- 250 machine cycles (0.375 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/14/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CLEAR_OLD_CRCS:
   mov   A, #num_regions      ; total number of regions
   jz    CLEAR_OLD_CRCS_99    ; no regions to clear
   
   mov   R0, A                ; initialize loop counter
   mov   DPTR, #CRC_REGIONS+8 ; 'previous CRC' of first region
CLEAR_OLD_CRCS_10:
   clr   A
   movx  @DPTR, A             ; clear LSB of 'previous CRC'
   inc   DPTR
   movx  @DPTR, A             ; clear MSB of 'previous CRC'
   ; advance pointer to next region (next element in the table)
   mov   A, #sz_regions-1     ; '-1' because we already incremented DPTR
   add   A, DPL
   mov   DPL, A
   mov   A, DPH
   addc  A, #0                ; propagate the carry
   mov   DPH, A
   djnz  R0, CLEAR_OLD_CRCS_10  ; loop until done

CLEAR_OLD_CRCS_99:
   ret                      ; CLEAR_OLD_CRCS
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- TDC_A  &  TDC_B
; Purpose -- Write 16 bits to either TDC 'A' or 'B'.
; Inputs -- Data to send in R0:R1.  Routine sends MSb first.
; Outputs -- 16 bits sent out appropriate 8051 port pin.
; Timing -- 205 machine cycles (0.308 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/21/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
TDC_A:
   clr   TDC_CLOCK          ; clock low (data latched on low-to-high)
   clr   TDC_B_ENABLE       ; ensure TDC 'B' is disabled
   setb  TDC_A_ENABLE       ; enable TDC 'A'
   sjmp  TDC_WRITE
TDC_B:
   clr   TDC_CLOCK          ; clock low (data latched on low-to-high)
   clr   TDC_A_ENABLE       ; ensure TDC 'A' is disabled
   setb  TDC_B_ENABLE       ; enable TDC 'B'
TDC_WRITE:
   mov   A, R0              ; MSB into ACC for shifting out
   mov   R0, #16            ; 16 bits to shift out
   ; we will rotate the 16-bit value held in ACC:R1 left 16 times
TDC_WRITE_10:
   xch   A, R1              ; put LSB in ACC (and hold MSB in R1)
   rlc   A                  ; put MS-bit of LSB into carry
   xch   A, R1              ; swap bytes back again
   rlc   A                  ; put carry into LS-bit of MSB and put next output
                            ; bit (MSB's MS-bit) into carry.
   mov   TDC_DATA, C        ; write Carry (the data bit) out port pin
   setb  TDC_CLOCK          ; Latch bit into ACTEL device (the TDC)
   nop                      ; remain high for 1.5 us
   nop
   clr   TDC_CLOCK          ; clock low in preparation for next cycle
   djnz  R0, TDC_WRITE_10   ; loop through 16 bits

TDC_WRITE_99:
   clr   TDC_A_ENABLE       ; disable TDC 'A'
   clr   TDC_B_ENABLE       ; disable TDC 'B'
   ret                      ; TDC_A  &  TDC_B
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- MEMORY_MONITORS
; Purpose -- Read the 8 memory monitor locations and store the 8 byte values.
;            (These will later be copied into the housekeeping packet.)
; Inputs -- 8 memory monitor address in parameter 'mLFDMADD' (little-endian).
;           Bitfield parameter 'mLFDMDATA' determines whether a particular
;           memory read is DATA/XDATA ('0') or CODE ('1').
;           Bitfield parameter 'mLFDMEXTR' further differentiates the data
;           source as EXTERNAL ('0') or INTERNAL ('1').
;           Bit 0 of the bitfield parameters corresponds to the first address
;           found in 'mLFDMADD'.
; Outputs -- 8 byte values are written to 'mLFDMONS' (which will later be
;            written to the housekeeping packet).
; Registers -- Bank 2
; Timing -- all 8 in CODE space --  496 machine cycles (0.744 msec @ 8-MHz)
;           all 8 in DATA space --  536 machine cycles (0.804 msec @ 8-MHz)
;           all 8 in XDATA space -- 560 machine cycles (0.840 msec @ 8-MHz)
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
MEMORY_MONITORS:
   push  PSW
   mov   PSW, #(2 SHL 3)    ; use register bank 2

   mov   R7, #8             ; loop counter -- number of addresses to monitor
   mov   DPTR, #mLFDMADD    ; starting location of table of addresses
   mov   R6, DPH            ; save current memory address in R6:R5
   mov   R5, DPL
   mov   DPTR, #mLFDMONS    ; starting location of data values to store
   mov   R4, DPH            ; save current memory address in R4:R3
   mov   R3, DPL
   mov   R2, #1             ; mask bit (begins with bit 0)

MEMORY_MONITORS_00:
   mov   DPH, R6            ; pointer to address
   mov   DPL, R5
   movx  A, @DPTR
   mov   tempDPTR+1, A      ; LSB of address to read
   inc   DPTR
   movx  A, @DPTR
   mov   tempDPTR+0, A      ; MSB of address to read
   mov   DPTR, #mLFDMDATA   ; bit fields:  0=CODE, 1=RAM (DATA or XDATA)
   movx  A, @DPTR
   anl   A, R2              ; mask off the word
   jnz   MEMORY_MONITORS_50 ; a CODE space address
   
MEMORY_MONITORS_10:
   ; Memory monitor points to RAM space (either DATA or XDATA) ...
   mov   DPTR, #mLFDMEXTR   ; bit fields:  0=XDATA, 1=DATA
   movx  A, @DPTR
   anl   A, R2              ; mask off the word
   jz    MEMORY_MONITORS_30 ; an XDATA address
   
MEMORY_MONITORS_20:
   ; Memory monitor points to DATA space (i.e. internal) ...
   mov   R1, tempDPTR+1     ; LSB is whole address of internal memory
   mov   A, @R1             ; read internal memory
   mov   DPH, R4            ; recall memory write address
   mov   DPL, R3
   movx  @DPTR, A           ; store the value
   sjmp  MEMORY_MONITORS_70 ; increment pointers, etc.
   
MEMORY_MONITORS_30:
   ; Memory monitor points to XDATA space ...
   mov   DPH, tempDPTR+0    ; recall memory read address
   mov   DPL, tempDPTR+1
   movx  A, @DPTR           ; read memory
   mov   DPH, R4            ; recall memory write address
   mov   DPL, R3
   movx  @DPTR, A           ; store the value
   sjmp  MEMORY_MONITORS_70 ; increment pointers, etc.

MEMORY_MONITORS_50:
   ; Memory monitor points to CODE space ...
   mov   DPH, tempDPTR+0    ; recall memory read address
   mov   DPL, tempDPTR+1
   clr   A
   movc  A, @A+DPTR         ; read memory
   mov   DPH, R4            ; recall memory write address
   mov   DPL, R3
   movx  @DPTR, A           ; store the value

MEMORY_MONITORS_70:
   mov   DPH, R4            ; advance storage pointer by 1 byte
   mov   DPL, R3
   inc   DPTR
   mov   R4, DPH
   mov   R3, DPL
   mov   DPH, R6            ; advance memory address pointer by 2 bytes
   mov   DPL, R5
   inc   DPTR
   inc   DPTR
   mov   R6, DPH
   mov   R5, DPL
   mov   A, R2              ; rotate mask byte by 1
   rl    A
   mov   R2, A              ; store new mask byte
   djnz  R7, MEMORY_MONITORS_00

MEMORY_MONITORS_99:
   pop   PSW                ; restore previous register bank in use
   ret                      ; MEMORY_MONITORS
;-----------------------------------------------------------------------------

;*****************************************************************************
; Subroutine -- XDATA_TLU_WORD
; Purpose -- Performs a table look-up for word-sized (16-bit) values from
;            XDATA
; Inputs -- DPTR points to base of structure, ACC is index into the table.
;           Accumulator must be in the range:  0 <= ACC <= 7Fh.  Table values
;           are assumed to be BIG-ENDIAN.  That is, the first value read is
;           the MSB and the next value (after incrementing DPTR) is the LSB.
; Outputs -- B:A holds the 2-byte value (B is MSB).
; Testing -- Tested by simulation and working, 9/12/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
XDATA_TLU_WORD:
   push  DPH                ; store DPTR
   push  DPL
   anl   A, #7Fh            ; mask of msb just in case it was '1'
   rl    A                  ; multiply by 2 (this is a WORD offset)
   add   A, DPL
   mov   DPL, A
   mov   A, DPH
   addc  A, #0
   mov   DPH, A
   movx  A, @DPTR
   mov   B, A               ; B = MSB of table look-up value
   inc   DPTR
   movx  A, @DPTR           ; A = LSB
   pop   DPL                ; restore DPTR
   pop   DPH
XDATA_TLU_WORD_99:
   ret                      ; XDATA_TLU_WORD
;-----------------------------------------------------------------------------   

;*****************************************************************************
; Subroutine -- CODE_TLU_WORD
; Purpose -- Performs a table look-up for word-sized (16-bit) values from
;            CODE
; Inputs -- DPTR points to base of structure, ACC is index into the table.
;           Accumulator must be in the range:  0 <= ACC <= 7Fh.  Table values
;           are assumed to be BIG-ENDIAN.  That is, the first value read is
;           the MSB and the next value (after incrementing DPTR) is the LSB.
; Outputs -- B:A holds the 2-byte value (B is MSB).
; Testing -- Tested by simulation and working, 9/29/00
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
CODE_TLU_WORD:
   push  DPH                ; store DPTR
   push  DPL
   anl   A, #7Fh            ; mask of msb just in case it was '1'
   rl    A                  ; multiply by 2 (this is a WORD offset)
   add   A, DPL
   mov   DPL, A
   mov   A, DPH
   addc  A, #0
   mov   DPH, A
   clr   A                  ; prepare for movc command
   movc  A, @A+DPTR
   mov   B, A               ; B = MSB of table look-up value
   inc   DPTR
   clr   A                  ; prepare for movc command
   movc  A, @A+DPTR         ; A = LSB
   pop   DPL                ; restore DPTR
   pop   DPH
CODE_TLU_WORD_99:
   ret                      ; CODE_TLU_WORD
;-----------------------------------------------------------------------------   

;*****************************************************************************
; 'COS_BOOT_TBL' is a table of function addresses.  When a command opcode is
; received, the address of the corresponding routine is looked up in this
; table and then a jump is performed.
; NOTE:  Routine 'CMD_ILLEGAL' handles commands that are illegal for both Boot
;    and Operate modes.  Routine 'CMD_OPERATE' handles those which are legal
;    for Operage mode but illegal for Boot mode.
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
COS_BOOT_TBL:
   DW   DCE_LFDNOOP   ; 00H *No Operation
   DW   DCE_LFDMADDR  ; 01H  index,addr,bank
   DW   DCE_LFDCRC    ; 02H  Inserts specific CRC region in background
   DW   DCE_LFDCOPY   ; 03H  Copy memory-to-memory on command
   DW   CMD_ILLEGAL   ; 04H  No Operation, undefined command
   DW   CMD_OPERATE   ; 05H *SET BITS FOR TDC-A START (BEGIN) CFD WALK ADJ
   DW   CMD_ILLEGAL   ; 06H *SET BITS FOR TDC-B START (BEGIN) CFD WALK ADJ
   DW   CMD_ILLEGAL   ; 07H  SET BITS FOR TDC-A CHARGE BASELINE INSPECTOR WINDOW
   DW   CMD_ILLEGAL   ; 08H  SET BITS FOR TDC-B CHARGE BASELINE INSPECTOR WINDOW
   DW   CMD_OPERATE   ; 09H *SET BITS FOR TDC-A STOP (CFD) WALK ADJUSTMENT
   DW   CMD_ILLEGAL   ; 0AH *SET BITS FOR TDC-B STOP (CFD) WALK ADJUSTMENT
   DW   CMD_ILLEGAL   ; 0BH  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 0CH  No Operation, undefined command
   DW   CMD_OPERATE   ; 0DH *set max setting for HV A (DAC w/ clamping diode)
   DW   CMD_ILLEGAL   ; 0EH  set max setting for HV B (DAC w/ clamping diode)
   DW   CMD_ILLEGAL   ; 0FH  Mask Box image operation (fast)
   DW   CMD_ILLEGAL   ; 10H  Mask Box image operation (fast)
   DW   CMD_ILLEGAL   ; 11H  Mask Box image operation (fast)
   DW   CMD_ILLEGAL   ; 12H  Mask Box image operation (fast)
   DW   CMD_ILLEGAL   ; 13H  Mask Box image operation (fast)
   DW   CMD_ILLEGAL   ; 14H  Mask Box image operation (fast)
   DW   CMD_OPERATE   ; 15H *TDC-A LOWER CHARGE (Y) THRESHOLD
   DW   CMD_ILLEGAL   ; 16H  SET BITS FOR TDC-B LOWER CHARGE (Y) THRESHOLD
   DW   CMD_OPERATE   ; 17H *TDC-A LOWER TIME (X) THRESHOLD GDGTTX
   DW   CMD_ILLEGAL   ; 18H *TDC-B LOWER TIME (Y) THRESHOLD
   DW   CMD_ILLEGAL   ; 19H  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 1AH  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 1BH  No Operation, undefined command
   DW   CMD_OPERATE   ; 1Ch  Overrides Endswitches
   DW   CMD_ILLEGAL   ; 1DH  TURNS ON_OFF STIM LAMP
   DW   CMD_OPERATE   ; 1Eh  ; (STIM_ON_OFF_A) TDC-A STIMULUS CONTROL
   DW   CMD_ILLEGAL   ; 1FH  TURNS TDC-B STIMULUS GENERATOR ON_OFF
   DW   CMD_ILLEGAL   ; 20H  SETS STIMULUS GENERATOR A FOR SWEEPING AMPLITUDES
   DW   CMD_ILLEGAL   ; 21H  SETS STIMULUS GENERATOR B FOR SWEEPING AMPLITUDES
   DW   CMD_OPERATE   ; 22H *SET BITS FOR TDC-A UPPER CHARGE (Y) THRESHOLD
   DW   CMD_ILLEGAL   ; 23H  SET BITS FOR TDC-B UPPER CHARGE (Y) THRESHOLD
   DW   CMD_OPERATE   ; 24H  SET BITS FOR TDC-A TIME (X) IMAGE POSITION OFFSET
   DW   CMD_ILLEGAL   ; 25H  SET BITS FOR TDC-B TIME (X) IMAGE POSITION OFFSET
   DW   CMD_OPERATE   ; 26H  SET BITS FOR TDC-A TIME (X) SCALE FACTOR
   DW   CMD_ILLEGAL   ; 27H  SET BITS FOR TDC-B TIME (X) SCALE FACTOR
   DW   CMD_ILLEGAL   ; 28H  set observe mode (observing, special ops, safe,..
   DW   CMD_ILLEGAL   ; 29H  set data format on RS-422{  (for calibration)}
   DW   CMD_ILLEGAL   ; 2AH  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 2BH *Enter test mode. (dump pseudo telemetry ...)
   DW   CMD_ILLEGAL   ; 2CH *Check a value (returned as a monitor)
   DW   DCE_LFDUPLOD  ; 2DH  Loads memory with Upload Data block
   DW   DCE_LFDDNLOD  ; 2EH *Dumps memory block: Startaddr, Length, DataCRC32
   DW   CMD_ILLEGAL   ; 2FH   spare
   DW   CMD_ILLEGAL   ; 30H  LOADS IMAGE MASKS FOR SEGMENT A
   DW   CMD_ILLEGAL   ; 31H  LOADS IMAGE MASKS FOR SEGMENT B
   DW   CMD_ILLEGAL   ; 32H  Dumps mask A: Startaddr, Endaddr, DataCRC32
   DW   CMD_ILLEGAL   ; 33H  Dumps mask B: Startaddr, Endaddr, DataCRC32
   DW   CMD_ILLEGAL   ; 34H *SAA COUNT LIMIT FOR SEGMENT A AUTO HV REDUCE
   DW   CMD_ILLEGAL   ; 35H  SAA COUNT LIMIT FOR SEGMENT B AUTO HV REDUCE
   DW   CMD_OPERATE   ; 36H *SAA HV REDUCED VALUE SEGMENT A
   DW   CMD_ILLEGAL   ; 37H  SAA HV REDUCED VALUE SEGMENT B
   DW   CMD_ILLEGAL   ; 38H  SAA HV AUTO SHUTDOWN OVER-RIDE SEGMENT A
   DW   CMD_ILLEGAL   ; 39H  SAA HV AUTO SHUTDOWN OVER-RIDE SEGMENT B
   DW   CMD_OPERATE   ; 3AH *SAA INTEGRATION TIME FOR SEGMENT A AUTO HV REDUCE
   DW   CMD_ILLEGAL   ; 3BH  SAA INTEGRATION TIME FOR SEGMENT B AUTO HV REDUCE
   DW   CMD_ILLEGAL   ; 3CH *SAA SEGMENT AUTO HV REDUCE OFF PERIOD
   DW   CMD_ILLEGAL   ; 3DH   spare
   DW   CMD_ILLEGAL   ; 3EH   spare
   DW   CMD_ILLEGAL   ; 3FH   spare
   DW   CMD_OPERATE   ; 40H *switches HV BIAS supplies on after writing 0 to the DA
   DW   CMD_OPERATE   ; 41H  TURNS ON_OFF ACTUATOR 1
   DW   CMD_OPERATE   ; 42H  TURNS ON_OFF ACTUATOR 2
   DW   CMD_OPERATE   ; 43H  ARMS 28V ACTUATOR PWR SWITCH
   DW   CMD_OPERATE   ; 44H  Turns ON_OFF 28V AUX OSCILLATOR for Motor, Actuator, I
   DW   CMD_OPERATE   ; 45H  Actuator reset sequence
   DW   CMD_OPERATE   ; 46H *LFHVNMA Nom value for HV A
   DW   CMD_ILLEGAL   ; 47H  LFHVNMB Nom value for HV B
   DW   CMD_ILLEGAL   ; 48H  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 49H  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 4AH *CRP_OVR CRP override
   DW   CMD_OPERATE   ; 4BH *SET BITS FOR MCP-A HV BIAS
   DW   CMD_ILLEGAL   ; 4CH  SET BITS FOR MCP-B HV BIAS
   DW   CMD_ILLEGAL   ; 4DH  No Operation, undefined command
   DW   CMD_OPERATE   ; 4EH *HV_RAMP HV Ramp time constant
   DW   CMD_OPERATE   ; 4FH  switches MCP grid supply on
   DW   CMD_OPERATE   ; 50H *Set HV current limit
   DW   CMD_OPERATE   ; 51H *Enables HV_ON and HVQ_ON (in the DPU = software)
   DW   CMD_ILLEGAL   ; 52H  TURNS ON ION PUMP #1
   DW   CMD_ILLEGAL   ; 53H  TURNS ON ION PUMP #2
   DW   CMD_ILLEGAL   ; 54H  ENABLES ION PUMPS #1& #2
   DW   CMD_ILLEGAL   ; 55H *TDC Begin Y interpolator offset Y_SHIFT_DAC
   DW   CMD_ILLEGAL   ; 56H *TDC End Y interpolator  offset Y_SHIFT_DAE
   DW   CMD_ILLEGAL   ; 57H *TDC Begin Y interpolator scale Y_SCALE_DAC
   DW   CMD_ILLEGAL   ; 58H *TDC End Y interpolator scale Y_SCALE_DAE
   DW   CMD_OPERATE   ; 59H  Sets motor bridge switches (dir.) avoid M1=M2=1
   DW   CMD_OPERATE   ; 5AH  ENABLES MOTOR BRIDGE SW. (may start the motor)
   DW   CMD_OPERATE   ; 5BH  Allows MOTOR_ENA (in the DPU)
   DW   CMD_OPERATE   ; 5CH  Set door current limit
   DW   CMD_ILLEGAL   ; 5DH  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 5EH  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 5FH  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 60H  Mask Box image operation
   DW   CMD_ILLEGAL   ; 61H *TDC Begin X interpolator  offset [#DGBOFFX]
   DW   CMD_ILLEGAL   ; 62H *TDC End X interpolator  offset [#DGEOFFX]
   DW   CMD_ILLEGAL   ; 63H *TDC Begin X interpolator scale [#DGBSCLX]
   DW   CMD_ILLEGAL   ; 64H *TDC End X interpolator scale [#DGESCLX]
   DW   CMD_ILLEGAL   ; 65H  Mask Box image operation
   DW   CMD_OPERATE   ; 66H *Trigger shutdown, no auto restore
   DW   CMD_ILLEGAL   ; 67H *Restore HV saved during shutdown
   DW   CMD_ILLEGAL   ; 68H  Set SAA shutdown values to 80% of HV
   DW   CMD_ILLEGAL   ; 69H  Don't expect anymore commands, no reset
   DW   DCE_LFDGOTO   ; 6AH *Jump to specified address and execute code there
   DW   CMD_ILLEGAL   ; 6BH *CRP enable/disable
   DW   CMD_ILLEGAL   ; 6CH  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 6DH  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 6EH *day - Restore HV saved during shutdown
   DW   CMD_ILLEGAL   ; 6FH *night - Trigger shutdown, no auto restore
   DW   DCE_LFDRSTP   ; 70H *power-on reset
   DW   DCE_LFDWDOG   ; 71H *enable/disable watchdog
   DW   CMD_ILLEGAL   ; 72H  No Operation, undefined command
   DW   DCE_LFDJMPCS  ; 73H *jump code
   DW   DCE_LFDDIAGC  ; 74H *clear diags
   DW   DCE_LFDRSTW   ; 75H *watchdog reset
   DW   CMD_ILLEGAL   ; 76H *Enter Idle state, HV off [#DSBITS5]
   DW   CMD_ILLEGAL   ; 77H  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 78H  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 79H  No Operation, undefined command
   DW   CMD_ILLEGAL   ; 7AH  Toggle reset line to counter actel
   DW   CMD_ILLEGAL   ; 7BH  TDC debug dbb 4-14-00
   DW   CMD_ILLEGAL   ; 7CH  Reserved for J&T
   DW   CMD_ILLEGAL   ; 7DH *Set PROM bit according to parameter
   DW   CMD_ILLEGAL   ; 7EH  dbb 4-1-00
   DW   DCE_LFDHKREQ  ; 7FH *Send full copy of Housekeeping Memory
;-----------------------------------------------------------------------------





$if INCLUDE_TEST_CODE
;*****************************************************************************
;*****************************************************************************
;*****************************************************************************
; Test code is placed in its own segment well beyond any boot code ...
   CSEG  at  TESTCODE_SEG
;*****************************************************************************
;*****************************************************************************
;*****************************************************************************

;-----------------------------------------------------------------------------
; Write byte 'value' to XDATA memory location 'dest' ...
WRITE_BYTE MACRO dest, value
$SAVE
$NOLIST
   mov   DPTR, #dest
   mov   A, #LOW(value)
   movx  @DPTR, A
$RESTORE
ENDM
;-----------------------------------------------------------------------------
; Write word 'value' to XDATA memory location 'dest' with MSB at lower
; address (big-endian) ...
WRITE_BIG_END MACRO dest, value
$SAVE
$NOLIST
   mov   DPTR, #dest
   mov   A, #HIGH(value)
   movx  @DPTR, A
   inc   DPTR
   mov   A, #LOW(value)
   movx  @DPTR, A
$RESTORE
ENDM
;-----------------------------------------------------------------------------
; Write word 'value' to XDATA memory location 'dest' with LSB at lower
; address (little-endian) ...
WRITE_LITTLE_END MACRO dest, value
$SAVE
$NOLIST
   mov   DPTR, #dest
   mov   A, #LOW(value)
   movx  @DPTR, A
   inc   DPTR
   mov   A, #HIGH(value)
   movx  @DPTR, A
$RESTORE
ENDM
;-----------------------------------------------------------------------------
; This macro puts command parameters directly into the command receipt buffer,
; thereby avoiding the overhead of the ISR call.  It also sets the 'RXD_FLAG',
; to tell the main routine that a command has been received.
TEST_MAKE_COMMAND MACRO  cmd, seq, p1, p2, p3, p4, p5
$SAVE
$NOLIST
cmd1   equ   ((LOW(cmd) SHL 8) OR LOW(cmd))
addr   equ   RS17

   mov   DPTR, #addr
   IRP  PARAM, <cmd1,seq,p1,p2,p3,p4,p5>
      mov   A, #LOW(PARAM)
      movx  @DPTR, A
      inc   DPTR
      mov   A, #HIGH(PARAM)
      movx  @DPTR, A
      inc   DPTR
      mov   A, #LOW(PARAM) XOR 0FFh
      movx  @DPTR, A
      inc   DPTR
      mov   A, #HIGH(PARAM) XOR 0FFh
      movx  @DPTR, A
      inc   DPTR
   ENDM      
   setb  RXD_FLAG           ; signal main routine that command received

$RESTORE
ENDM
;-----------------------------------------------------------------------------
; This macro puts command parameters in the receive registers and calls
; the command receive ISR to process them in normal fashion ...
TEST_MAKE_COMMAND1 MACRO  cmd, seq, p1, p2, p3, p4, p5
$SAVE
$NOLIST
cmd_buffer   equ   0440h
addr         equ   cmd_buffer+4

   IRP  PARAM, <seq,p1,p2,p3,p4,p5>
      mov   DPTR, #RXD_HARDWARE+300h   ; put command MSB here
      mov   A, #HIGH(addr)
      movx  @DPTR, A
      dec   DPH
      mov   A, #LOW(addr)
      movx  @DPTR, A
      dec   DPH
      mov   A, #HIGH(PARAM)
      movx  @DPTR, A
      dec   DPH
      mov   A, #LOW(PARAM)
      movx  @DPTR, A
      lcall ISR_Pri_Cmd_Received
addr   equ   addr+2
      mov   DPTR, #RXD_HARDWARE+300h   ; put command MSB here
      mov   A, #HIGH(addr)
      movx  @DPTR, A
      dec   DPH
      mov   A, #LOW(addr)
      movx  @DPTR, A
      dec   DPH
      mov   A, #HIGH(PARAM) XOR 0FFh
      movx  @DPTR, A
      dec   DPH
      mov   A, #LOW(PARAM) XOR 0FFh
      movx  @DPTR, A
      lcall ISR_Pri_Cmd_Received
addr   equ   addr+2
   ENDM

   ; Finally insert the command commplement and then the command ...
addr   equ   cmd_buffer+2
   mov   DPTR, #RXD_HARDWARE+300h   ; put command MSB here
   mov   A, #HIGH(addr)
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(addr)
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(cmd) XOR 0FFh
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(cmd) XOR 0FFh
   movx  @DPTR, A
   lcall ISR_Pri_Cmd_Received
addr   equ   cmd_buffer
   mov   DPTR, #RXD_HARDWARE+300h   ; put command MSB here
   mov   A, #HIGH(addr)
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(addr)
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(cmd)
   movx  @DPTR, A
   dec   DPH
   mov   A, #LOW(cmd)
   movx  @DPTR, A
   lcall ISR_Pri_Cmd_Received

$RESTORE
ENDM
;-----------------------------------------------------------------------------
; This macro creates a command sequence in CODE space for later copying
; to the command receipt buffer.
TEST_MAKE_COMMAND2 MACRO  cmd, seq, p1, p2, p3, p4, p5
cmd1     equ   ((LOW(cmd) SHL 8) OR LOW(cmd))
   IRP  PARAM, <cmd1,seq_num,p1,p2,p3,p4,p5>
      DB   LOW(PARAM), HIGH(PARAM), LOW(NOT PARAM), HIGH(NOT PARAM)
   ENDM      
seq_num   equ   seq_num+1
ENDM
;-----------------------------------------------------------------------------
; This macro creates a command in which the opcode is not in both upper and
; lower halves of the command word
TEST_MAKE_BAD_CMD MACRO
   DB   0, 80h, 0FFh, 07Fh
   DB   LOW(seq_num), HIGH(seq_num), LOW(NOT seq_num), HIGH(NOT seq_num)
   DB   0,   0, 0FFh, 0FFh
   DB   0,   0, 0FFh, 0FFh
   DB   0,   0, 0FFh, 0FFh
   DB   0,   0, 0FFh, 0FFh
   DB   0,   0, 0FFh, 0FFh
seq_num   equ   seq_num+1
ENDM
;-----------------------------------------------------------------------------
; This macro creates a command in which a parameter does not have a proper
; complement
TEST_MAKE_BAD_PARAM MACRO
   DB   80h, 80h, 07Fh, 07Fh
   DB   LOW(seq_num), HIGH(seq_num), LOW(NOT seq_num), HIGH(NOT seq_num)
   DB     0,   0, 0FFh, 0FFh
   DB     0,   0, 0FFh, 0FFh
   DB     0,   0,    0, 0FFh
   DB     0,   0, 0FFh, 0FFh
   DB     0,   0, 0FFh, 0FFh
seq_num   equ   seq_num+1
ENDM
;-----------------------------------------------------------------------------

;*****************************************************************************
; 'TEST_MEMORY_MONITORS' inserts data into the memory monitor structure
; and verifies that 
;*****************************************************************************
; Test MEMORY_MONITORS
FILL_TO_BOUNDARY   16, 0FFh
TEST_MEMORY_MONITORS:
   WRITE_BYTE  mLFDMDATA, 10101010b
   WRITE_BYTE  mLFDMEXTR, 00001111b
   WRITE_LITTLE_END  mLFDMADD+0, 6000h+128+0
   WRITE_LITTLE_END  mLFDMADD+2, 6000h+128+1
   WRITE_LITTLE_END  mLFDMADD+4, 6000h+128+2
   WRITE_LITTLE_END  mLFDMADD+6, 6000h+128+3
   WRITE_LITTLE_END  mLFDMADD+8, 6000h+128+4
   WRITE_LITTLE_END  mLFDMADD+10, 6000h+128+5
   WRITE_LITTLE_END  mLFDMADD+12, 6000h+128+6
   WRITE_LITTLE_END  mLFDMADD+14, 6000h+128+7
   WRITE_BYTE  6000h+128+0, 5
   WRITE_BYTE  6000h+128+1, 10
   WRITE_BYTE  6000h+128+2, 15
   WRITE_BYTE  6000h+128+3, 20
   WRITE_BYTE  6000h+128+4, 25
   WRITE_BYTE  6000h+128+5, 30
   WRITE_BYTE  6000h+128+6, 35
   WRITE_BYTE  6000h+128+7, 40
   mov   R0, #128+0
   mov   @R0, #5
   mov   R0, #128+2
   mov   @R0, #15
;   lcall MEMORY_MONITORS
   ret
;-----------------------------------------------------------------------------

;*****************************************************************************
; Test command jump table ...
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
TEST_COMMAND_VECTOR:
   mov   HKPBUFF, #0        ; Temporary use
BOOT_MAIN_00:
   mov   A, HKPBUFF
   lcall COMMAND_VECTOR
   inc   HKPBUFF
   mov   A, HKPBUFF
   jnz   BOOT_MAIN_00
   ret
;-----------------------------------------------------------------------------

;*****************************************************************************
; Create command packet 80000000h to test 'jump to power-on reset' ...
;*****************************************************************************
FILL_TO_BOUNDARY   16, 0FFh
TEST_CMD80000000:
   mov   A, #80h
   mov   DPTR, #RXD_HARDWARE + 300h   ; Primary command input channel
   movx  @DPTR, A
   mov   A, #0
   dec   DPH
   movx  @DPTR, A
   dec   DPH
   movx  @DPTR, A
   dec   DPH
   movx  @DPTR, A
   lcall ISR_Pri_Cmd_Received
   ret
;-----------------------------------------------------------------------------

;*****************************************************************************
; 'test_timing' calls all the specified commands so timing can be computed.
; One should reset the timer (in debug mode) prior to each call.
;*****************************************************************************
; Call this subroutine on the occurrence of a timer TOCK (1-second timeout).
FILL_TO_BOUNDARY   16, 0FFh
test_timing:
   lcall ISR_Pri_Cmd_Received  ; Boot mode ISR 03
   lcall ISR_Timer0            ; Boot mode ISR 0B
   lcall ISR_Sec_Cmd_Received  ; Boot mode ISR 13
   lcall ISR_Timer1            ; Boot mode ISR 1B

   lcall CLEAR_HKP_PKT

   ret                      ; test_timing
;-----------------------------------------------------------------------------

;*****************************************************************************
; 'test_commands' uses the 1-second timer to select the next function from
; a pre-defined list of functions.  This function is called on the 1-second
; 'tock' in the main executive loop.
;*****************************************************************************
; Call this subroutine on the occurrence of a timer TOCK (1-second timeout).
FILL_TO_BOUNDARY   16, 0FFh
test_commands:

   ; Prepare registers for MEM_COPY routine.
   ; Put source address in R3:R2 ...
   mov   DPTR, #mLFCTIME
   movx  A, @DPTR           ; 1-second timer's LSB determines next command
   mov   B, #(cmd_sequence_1 - cmd_sequence)  ; spacing between commands
   mul   AB
   add   A, #LOW(cmd_sequence)
   mov   R2, A
   mov   A, B
   addc  A, #HIGH(cmd_sequence)
   mov   R3, A
   ; Put destination address in R5:R4 (command-received buffer) ...
   mov   R5, #HIGH(RS17)
   mov   R4, #LOW(RS17)
   ; Number of bytes to copy in R0:R7 (28 bytes for command packets) ...
   mov   R0, #HIGH(28)
   mov   R7, #LOW(28)
   ; Source of move is CODE space ...
   setb  SRC_FLAG           ; 1 = copy from CODE space
   lcall MEM_COPY           ; copy the command into the buffer
   setb  RXD_FLAG           ; flag a command received

test_commands_99:
   ret
;-----------------------------------------------------------------------------


; Create various command packets for testing.  These are copied to the
; command-received buffer once per second and RXD_FLAG is set true so
; the main executive loop will execute the command.
$SET (cmd_seq_num=1)

seq_num   equ   0   ; Macros increment this giving a rolling seq. number

cmd_sequence:
;00 -- not executed because first timer rollover is to '1'
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
cmd_sequence_1:   ; label used to find spacing
$if (cmd_seq_num=1)
;01
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
;04
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP

   TEST_MAKE_COMMAND2    081h, 0,     0, 2460h,      0, 0, 0   ; LFDMADDR
   TEST_MAKE_COMMAND2    081h, 0,     1, 2461h,      0, 0, 0   ; LFDMADDR
   TEST_MAKE_COMMAND2    081h, 0,     2, 24BAh,      0, 0, 0   ; LFDMADDR
   TEST_MAKE_COMMAND2    081h, 0,     3, 1A20h,      1, 0, 0   ; LFDMADDR
   TEST_MAKE_COMMAND2    081h, 0,     4, 1A21h,      1, 0, 0   ; LFDMADDR
   TEST_MAKE_COMMAND2    081h, 0,     5, 1A22h,      1, 0, 0   ; LFDMADDR
   TEST_MAKE_COMMAND2    081h, 0,     6, 0031h,      2, 0, 0   ; LFDMADDR
   TEST_MAKE_COMMAND2    081h, 0,     7, 0032h,      2, 0, 0   ; LFDMADDR

   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    082h, 0, 6000h, 0400h,      0, 0, 0   ; LFDCRC
;10
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    095h, 0,     0,     0,      0, 0, 0   ; error
   TEST_MAKE_COMMAND2    035h, 0,     0,     0,      0, 0, 0   ; error
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    0F4h, 0,     0,     0,      0, 0, 0   ; LFDDIAGC
;19
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    083h, 0,     0, 4000h,  0400h, 0, 0   ; LFDCOPY
;1D
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    0ADh, 0, 6000h, 0400h,      0, 0, 0   ; LFDUPLOD
   TEST_MAKE_COMMAND2    0ADh, 0, 6000h, 0400h,      0, 0, 0   ; LFDUPLOD
;22
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    0AEh, 0, 6000h, 0400h,      0, 0, 0   ; LFDDNLOD
;26
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    0EAh, 0, 0300h,     0,      0, 0, 0   ; LFDGOTO
;2A
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    0F1h, 0,     1,     0,      0, 0, 0   ; LFDWDOG ON
   TEST_MAKE_COMMAND2    0F1h, 0,     0,     0,      0, 0, 0   ; LFDWDOG OFF
;2F
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    0F3h, 0,     2,     0,      0, 0, 0   ; LFDJMPCS
;33
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    0F0h, 0,     0,     0,      0, 0, 0   ; LFDRSTP
;37
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    0F5h, 0,     0,     0,      0, 0, 0   ; LFDRSTW
;3B
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    085h, 0,     0,     0,      0, 0, 0   ; Operate Only
   TEST_MAKE_COMMAND2    084h, 0,     0,     0,      0, 0, 0   ; Illegal
;40
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_BAD_CMD
   TEST_MAKE_BAD_PARAM
;45   
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
$elseif (cmd_seq_num=2)
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    0ADh, 0, 6000h, 0400h,      0, 0, 0   ; LFDUPLOD
   TEST_MAKE_COMMAND2    0ADh, 0, 6000h, 0400h,      0, 0, 0   ; LFDUPLOD
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
$elseif (cmd_seq_num=3)
$elseif (cmd_seq_num=4)
$elseif (cmd_seq_num=5)
$elseif (cmd_seq_num=6)
$elseif (cmd_seq_num=7)
$elseif (cmd_seq_num=8)
$elseif (cmd_seq_num=9)
$else
   TEST_MAKE_COMMAND2    080h, 0,     0,     0,      0, 0, 0   ; LFDNOOP
   TEST_MAKE_COMMAND2    081h, 0,     0, 2460h,      0, 0, 0   ; LFDMADDR
   TEST_MAKE_COMMAND2    082h, 0, 6000h, 0400h,      0, 0, 0   ; LFDCRC
   TEST_MAKE_COMMAND2    083h, 0,     0, 4000h,  0400h, 0, 0   ; LFDCOPY
   TEST_MAKE_COMMAND2    0ADh, 0, 6000h, 0400h,      0, 0, 0   ; LFDUPLOD
   TEST_MAKE_COMMAND2    0AEh, 0, 6000h, 0400h,      0, 0, 0   ; LFDDNLOD
   TEST_MAKE_COMMAND2    0EAh, 0,     0,     0,      0, 0, 0   ; LFDGOTO
   TEST_MAKE_COMMAND2    0F0h, 0,     0,     0,      0, 0, 0   ; LFDRSTP
   TEST_MAKE_COMMAND2    0F1h, 0,     1,     0,      0, 0, 0   ; LFDWDOG ON
   TEST_MAKE_COMMAND2    0F1h, 0,     0,     0,      0, 0, 0   ; LFDWDOG OFF
   TEST_MAKE_COMMAND2    0F3h, 0,     2,     0,      0, 0, 0   ; LFDJMPCS
   TEST_MAKE_COMMAND2    0F4h, 0,     0,     0,      0, 0, 0   ; LFDDIAGC
   TEST_MAKE_COMMAND2    0F5h, 0,     0,     0,      0, 0, 0   ; LFDRSTW
   TEST_MAKE_COMMAND2    0FFh, 0,     0,     0,      0, 0, 0   ; LFDHKREQ
   TEST_MAKE_COMMAND2    095h, 0,     0,     0,      0, 0, 0   ; error
   TEST_MAKE_COMMAND2    035h, 0,     0,     0,      0, 0, 0   ; error
$endif      ; (cmd_seq_num=??)
;-----------------------------------------------------------------------------

$endif      ; INCLUDE_TEST_CODE

END                         ; bootmain.a51



